try:
    import v20
except ImportError:
    raise Exception("Please use 'pip install v20' to trade using Oanda v20 API.")
import sys
import time
import datetime
import traceback
import numpy as np
import pandas as pd
from autotrader import utilities, AutoData
from autotrader.brokers.broker import AbstractBroker
from autotrader.brokers.broker_utils import BrokerUtils
from autotrader.brokers.trading import Order, IsolatedPosition, Position, Trade


class Broker(AbstractBroker):
    def __init__(self, oanda_config: dict, utils: BrokerUtils = None):
        """Create v20 context."""
        self.API = oanda_config["API"]
        self.ACCESS_TOKEN = oanda_config["ACCESS_TOKEN"]
        self.port = oanda_config["PORT"]
        self.ACCOUNT_ID = oanda_config["ACCOUNT_ID"]
        self.api = v20.Context(
            hostname=self.API, token=self.ACCESS_TOKEN, port=self.port
        )

        self.STREAM_API = "stream-fxpractice.oanda.com"
        self.stream = v20.Context(
            hostname=self.STREAM_API, token=self.ACCESS_TOKEN, port=self.port
        )
        self.open_positions = {}

        # Assign broker utilities
        if utils is None:
            utils = BrokerUtils()
        self._utils = utils

        # Create AutoData instance
        data_config = utilities.get_data_config(feed="oanda")
        self.autodata = AutoData(data_config)

    def __repr__(self):
        return "AutoTrader-Oanda Broker Interface"

    def __str__(self):
        return "AutoTrader-Oanda Broker Interface"

    def get_NAV(self) -> float:
        """Returns Net Asset Value of account."""
        self._check_connection()
        response = self.api.account.get(accountID=self.ACCOUNT_ID)
        return response.body["account"].NAV

    def get_balance(self) -> float:
        """Returns account balance."""
        self._check_connection()
        response = self.api.account.get(accountID=self.ACCOUNT_ID)
        return response.body["account"].balance

    def place_order(self, order: Order, **kwargs):
        """Submits order to broker."""
        self._check_connection()

        # Call order to set order time
        order()

        # Submit order
        if order.order_type == "market":
            response = self._place_market_order(order)
        elif order.order_type == "stop-limit":
            response = self._place_stop_limit_order(order)
        elif order.order_type == "limit":
            response = self._place_limit_order(order)
        elif order.order_type == "close":
            response = self._close_position(order.instrument)
        elif order.order_type == "modify":
            response = self._modify_trade(order)
        else:
            print("Order type not recognised.")

        # Check response
        # output = self._check_response(response)

        return response

    def get_orders(self, instrument=None, **kwargs) -> dict:
        """Get all pending orders in the account."""
        self._check_connection()
        response = self.api.order.list_pending(
            accountID=self.ACCOUNT_ID, instrument=instrument
        )
        oanda_pending_orders = response.body["orders"]
        orders = {}

        for order in oanda_pending_orders:
            if order.type != "TAKE_PROFIT" and order.type != "STOP_LOSS":
                new_order = {}
                new_order["id"] = order.id
                new_order["status"] = "open"
                new_order["order_type"] = order.type
                new_order["order_stop_price"] = order.price
                new_order["order_limit_price"] = order.price
                new_order["direction"] = np.sign(order.units)
                new_order["order_time"] = order.createTime
                new_order["instrument"] = order.instrument
                new_order["size"] = abs(order.units)
                new_order["order_price"] = order.price

                # Check for take profit
                if order.takeProfitOnFill is not None:
                    new_order["take_profit"] = order.takeProfitOnFill.price

                # Check for stop loss
                if order.stopLossOnFill is not None:
                    new_order["stop_loss"] = order.stopLossOnFill.price
                    new_order["stop_type"] = "limit"

                if instrument is not None and order.instrument == instrument:
                    orders[order.id] = Order._from_dict(new_order)
                elif instrument is None:
                    orders[order.id] = Order._from_dict(new_order)

        return orders

    def cancel_order(self, order_id: int, **kwargs) -> None:
        """Cancels pending order by ID."""
        self._check_connection()
        self.api.order.cancel(accountID=self.ACCOUNT_ID, orderSpecifier=str(order_id))

    def get_trades(self, instrument=None, **kwargs) -> list:
        """Returns the trades (fills) made by the account."""
        self._check_connection()
        response = self.api.trade.list_open(
            accountID=self.ACCOUNT_ID, state="CLOSED", instrument=instrument
        )
        oanda_trades = response.body["trades"]

        trades = {}
        for trade in oanda_trades:
            trade_dict = self._oanda_trade_to_dict(trade)

            native_trade = Trade(
                instrument=trade_dict["instrument"],
                order_price=None,
                order_time=None,
                order_type=None,
                size=float(trade_dict["size"]),
                fill_time=trade_dict["time_filled"],
                fill_price=float(trade_dict["fill_price"]),
                fill_direction=trade_dict["direction"],
                fee=trade_dict["fees"],
                id=trade_dict["id"],
            )

            trades[trade.id] = native_trade

        return trades

    def get_isolated_positions(self, instrument: str = None, **kwargs):
        """Returns isolated positions for the specified instrument.

        Parameters
        ----------
        instrument : str, optional
            The instrument to fetch trades under. The default is None.
        """
        self._check_connection()
        response = self.api.trade.list_open(accountID=self.ACCOUNT_ID)
        oanda_open_trades = response.body["trades"]

        open_trades = {}
        for trade in oanda_open_trades:
            new_trade = self._oanda_trade_to_dict(trade)

            # Filter by instrument
            if instrument is not None and trade.instrument in instrument:
                open_trades[trade.id] = IsolatedPosition(**new_trade)
            elif instrument is None:
                open_trades[trade.id] = IsolatedPosition(**new_trade)

        return open_trades

    @staticmethod
    def _oanda_trade_to_dict(trade) -> dict:
        """Converts an Oanda Trade object to a dictionary."""
        new_trade = {}
        related_orders = []
        new_trade["instrument"] = trade.instrument
        new_trade["time_filled"] = trade.openTime
        new_trade["fill_price"] = trade.price
        new_trade["size"] = abs(trade.currentUnits)
        new_trade["id"] = trade.id
        new_trade["direction"] = np.sign(trade.currentUnits)
        new_trade["margin_required"] = trade.marginUsed
        new_trade["unrealised_PL"] = trade.unrealizedPL
        new_trade["fees"] = trade.financing
        new_trade["status"] = trade.state.lower()

        # Check for take profit
        if trade.takeProfitOrder is not None:
            new_trade["take_profit"] = trade.takeProfitOrder.price
            related_orders.append(trade.takeProfitOrder.id)

        # Check for stop loss
        if trade.stopLossOrder is not None:
            new_trade["stop_loss"] = trade.stopLossOrder.price
            new_trade["stop_type"] = "limit"
            related_orders.append(trade.stopLossOrder.id)

        if related_orders is not None:
            new_trade["related_orders"] = related_orders

        return new_trade

    def get_trade_details(self, trade_ID: int):
        """Returns the details of the trade specified by trade_ID.
        WARNING: THIS METHOD HAS BEEN MADE REDUNDANT BY 'get_trades' METHOD.
        """
        raise DeprecationWarning(
            "This method is deprecated, and will "
            + "be removed in a future release. Please use the "
            + "get_trades method instead."
        )
        response = self.api.trade.list(accountID=self.ACCOUNT_ID, ids=int(trade_ID))
        trade = response.body["trades"][0]

        details = {
            "direction": int(np.sign(trade.currentUnits)),
            "order_time": datetime.datetime.strptime(
                trade.openTime[:-4], "%Y-%m-%dT%H:%M:%S.%f"
            ),
            "instrument": trade.instrument,
            "size": trade.currentUnits,
            "order_price": trade.price,
            "order_ID": trade.id,
            "time_filled": trade.openTime,
            "entry_price": trade.price,
            "unrealised_PL": trade.unrealizedPL,
            "margin_required": trade.marginUsed,
        }

        # Get associated trades
        related = []
        try:
            details["take_profit"] = trade.takeProfitOrder.price
            related.append(trade.takeProfitOrder.id)
        except:
            pass

        try:
            details["stop_loss"] = trade.stopLossOrder.price
            related.append(trade.stopLossOrder.id)
        except:
            pass
        details["related_orders"] = related

        # TODO - veryify functionality of below...
        return Trade(trade)

    def get_positions(self, instrument: str = None, **kwargs) -> dict:
        """Gets the current positions open on the account."""
        self._check_connection()
        response = self.api.position.list_open(accountID=self.ACCOUNT_ID)
        oanda_open_positions = response.body["positions"]
        open_positions = {}
        for position in oanda_open_positions:
            pos = {
                "instrument": position.instrument,
                "long_units": position.long.units,
                "long_PL": position.long.unrealizedPL,
                "long_margin": None,
                "short_units": position.short.units,
                "short_PL": position.short.unrealizedPL,
                "short_margin": None,
                "total_margin": position.marginUsed,
            }

            # fetch trade ID'strade_IDs
            trade_IDs = []
            if abs(pos["long_units"]) > 0:
                for ID in position.long.tradeIDs:
                    trade_IDs.append(ID)
            if abs(pos["short_units"]) > 0:
                for ID in position.short.tradeIDs:
                    trade_IDs.append(ID)

            pos["trade_IDs"] = trade_IDs

            if instrument is not None and position.instrument == instrument:
                open_positions[position.instrument] = Position(**pos)
            elif instrument is None:
                open_positions[position.instrument] = Position(**pos)

        return open_positions

    def get_position(self, instrument: str) -> Position:
        """Gets position from Oanda."""
        self._check_connection()
        response = self.api.position.get(
            instrument=instrument, accountID=self.ACCOUNT_ID
        )
        # TODO - convert to position
        return response.body["position"]

    def get_summary(self):
        """Returns account summary."""
        self._check_connection()
        # response = self.api.account.get(accountID=self.ACCOUNT_ID)
        response = self.api.account.summary(accountID=self.ACCOUNT_ID)
        return response

    def get_data(self, pair: str, period: int, interval: str) -> pd.DataFrame:
        self._check_connection()
        response = self.api.instrument.candles(
            pair, granularity=interval, count=period, dailyAlignment=0
        )
        data = self._utils.response_to_df(response)
        return data

    def check_trade_size(self, instrument: str, units: float) -> float:
        """Checks the requested trade size against the minimum trade size
        allowed for the currency pair.
        """
        response = self.api.account.instruments(
            accountID=self.ACCOUNT_ID, instruments=instrument
        )
        # minimum_units = response.body['instruments'][0].minimumTradeSize
        trade_unit_precision = response.body["instruments"][0].tradeUnitsPrecision
        return round(units, trade_unit_precision)

    def update_data(
        self, instrument: str, granularity: str, data: pd.DataFrame
    ) -> pd.DataFrame:
        """Attempts to construct the latest candle when there is a delay in the
        api feed.
        """
        self._check_connection()
        granularity_details = self.deconstruct_granularity(granularity)
        secs = granularity_details["seconds"]
        mins = granularity_details["minutes"]
        hrs = granularity_details["hours"]
        days = granularity_details["days"]

        small_granularity = self.get_reduced_granularity(granularity_details, 25)

        # Get data equivalent of last candle's granularity
        time_now = datetime.datetime.now()
        start_time = time_now - datetime.timedelta(
            seconds=secs, minutes=mins, hours=hrs, days=days
        )
        latest_data = self.get_historical_data(
            instrument, small_granularity, start_time.timestamp(), time_now.timestamp()
        )

        # Get latest price data
        latest_close = latest_data.Close.values[0]

        open_price = data.Close.values[-1]
        close_price = latest_close
        high_price = max(latest_data.High.values)
        low_price = min(latest_data.Low.values)
        last_time = data.index[-1]
        stripped_time = datetime.datetime.strptime(
            last_time.strftime("%Y-%m-%d %H:%M:%S%z"), "%Y-%m-%d %H:%M:%S%z"
        )
        new_time = stripped_time + datetime.timedelta(
            seconds=secs, minutes=mins, hours=hrs, days=days
        )

        new_candle = pd.DataFrame(
            {
                "Open": open_price,
                "High": high_price,
                "Low": low_price,
                "Close": close_price,
            },
            index=[new_time],
        )

        new_data = pd.concat([data, new_candle])

        return new_data

    def get_historical_data(self, instrument, interval, from_time, to_time):

        self._check_connection()

        response = self.api.instrument.candles(
            instrument, granularity=interval, fromTime=from_time, toTime=to_time
        )

        data = self._utils.response_to_df(response)

        return data

    def deconstruct_granularity(self, granularity: str):
        """Returns a dict with the seconds, minutes, hours and days
        corresponding to a granularity.
        """

        # Get letter to determine timeframe (eg. M)
        letter = granularity[0]

        # Get timeframe multiple (eg. 15)
        if len(granularity) > 1:
            number = float(granularity[1:])
        else:
            number = 1

        if letter == "S":
            seconds = number
            minutes = 0
            hours = 0
            days = 0

        elif letter == "M":
            seconds = 0
            minutes = number
            hours = 0
            days = 0

        elif letter == "H":
            seconds = 0
            minutes = 0
            hours = number
            days = 0

        else:
            seconds = 0
            minutes = 0
            hours = 0
            days = number

        granularity_details = {
            "seconds": seconds,
            "minutes": minutes,
            "hours": hours,
            "days": days,
        }

        return granularity_details

    def get_reduced_granularity(self, granularity_details, fraction):
        """Returns a candlestick granularity as a fraction of given granularity"""
        secs = granularity_details["seconds"]
        mins = granularity_details["minutes"]
        hrs = granularity_details["hours"]
        days = granularity_details["days"]

        total_seconds = secs + 60 * mins + 60 * 60 * hrs + 60 * 60 * 24 * days

        fractional_seconds = total_seconds / fraction

        seconds = fractional_seconds
        minutes = fractional_seconds / 60
        hours = fractional_seconds / (60 * 60)
        days = fractional_seconds / (60 * 60 * 24)

        if days > 1:
            letter = "D"
            number = 1
            reduced_granularity = letter

        elif hours > 1:
            base = 2
            letter = "H"
            number = base * round(hours / base)
            if number > 12:
                number = 12
            reduced_granularity = letter + str(number)

        elif minutes > 1:
            base = 15
            letter = "M"
            number = base * round(minutes / base)
            if number > 30:
                number = 30
            reduced_granularity = letter + str(number)

        else:
            base = 15
            letter = "S"
            number = base * round(seconds / base)
            if number > 30:
                number = 30
            reduced_granularity = letter + str(number)

        if reduced_granularity[1:] == "0":
            reduced_granularity = reduced_granularity[0] + "1"

        return reduced_granularity

    def get_pip_location(self, instrument: str):
        """Returns the pip location of the requested instrument."""
        response = self.api.account.instruments(self.ACCOUNT_ID, instruments=instrument)
        return response.body["instruments"][0].pipLocation

    def _check_connection(self) -> None:
        """Connects to Oanda v20 REST API. An initial call is performed to check
        for a timeout error.
        """
        # TODO - improve this - currently doubles the poll rate
        for atempt in range(10):
            try:
                # Attempt basic task to check connection
                self.api.account.get(accountID=self.ACCOUNT_ID)

            except BaseException:
                # Error has occurred
                ex_type, ex_value, ex_traceback = sys.exc_info()

                # Extract unformatter stack traces as tuples
                trace_back = traceback.extract_tb(ex_traceback)

                # Format stacktrace
                stack_trace = list()

                for trace in trace_back:
                    trade_string = "File : %s , Line : %d, " % (
                        trace[0],
                        trace[1],
                    ) + "Func.Name : %s, Message : %s" % (trace[2], trace[3])
                    stack_trace.append(trade_string)

                print("\nWARNING FROM OANDA API: The following exception was caught.")
                print(
                    "Time: {}".format(
                        datetime.datetime.now().strftime("%b %d %H:%M:%S")
                    )
                )
                print("Exception type : %s " % ex_type.__name__)
                print("Exception message : %s" % ex_value)
                print("Stack trace : %s" % stack_trace)
                print("  Attempting to reconnect to Oanda v20 API.")

                time.sleep(3)
                api = v20.Context(
                    hostname=self.API, token=self.ACCESS_TOKEN, port=self.port
                )
                self.api = api

            else:
                break

        else:
            print("FATAL: All attempts to connect to Oanda API have failed.")

    def _place_market_order(self, order: Order):
        """Places market order."""
        self._check_connection()
        stop_loss_order = self._get_stop_loss_order(order)
        take_profit_details = self._get_take_profit_details(order)

        # Check position size
        size = self.check_trade_size(order.instrument, order.size)

        response = self.api.order.market(
            accountID=self.ACCOUNT_ID,
            instrument=order.instrument,
            units=order.direction * size,
            takeProfitOnFill=take_profit_details,
            **stop_loss_order
        )

        return response

    def _place_stop_limit_order(self, order):
        """Places MarketIfTouchedOrder with Oanda.
        https://developer.oanda.com/rest-live-v20/order-df/
        """
        # TODO - this submits market if touched, options below
        ordertype = "MARKET_IF_TOUCHED"  # 'MARKET_IF_TOUCHED' # 'STOP', 'LIMIT'
        self._check_connection()

        stop_loss_order = self._get_stop_loss_order(order)
        take_profit_details = self._get_take_profit_details(order)

        # Check and correct order stop price
        price = self._check_precision(order.instrument, order.order_stop_price)
        trigger_condition = order.trigger_price
        size = self.check_trade_size(order.instrument, order.size)

        # Need to test cases when no stop/take is provided (as None type)
        response = self.api.order.market_if_touched(
            accountID=self.ACCOUNT_ID,
            instrument=order.instrument,
            units=order.direction * size,
            price=str(price),
            type=ordertype,
            takeProfitOnFill=take_profit_details,
            triggerCondition=trigger_condition,
            **stop_loss_order
        )
        return response

    def _place_stop_order(self, order: Order):
        """Places a stop order."""
        # TODO - implement this method
        self._check_connection()

        stop_loss_order = self._get_stop_loss_order(order)
        take_profit_details = self._get_take_profit_details(order)

        # Check and correct order stop price
        price = self._check_precision(order.instrument, order.order_stop_price)
        price_bound = self._check_precision(order.instrument, order.order_stop_price)

        trigger_condition = order.trigger_price
        size = self.check_trade_size(order.instrument, order.size)

        response = self.api.order.stop(
            accountID=self.ACCOUNT_ID,
            instrument=order.instrument,
            units=order.direction * size,
            price=str(price),
            priceBound=str(price_bound),
            triggerCondition=trigger_condition,
            takeProfitOnFill=take_profit_details,
            **stop_loss_order
        )
        return response

    def _place_limit_order(self, order: Order):
        """PLaces a limit order."""
        self._check_connection()

        stop_loss_order = self._get_stop_loss_order(order)
        take_profit_details = self._get_take_profit_details(order)

        # Check and correct order stop price
        price = self._check_precision(order.instrument, order.order_limit_price)

        trigger_condition = order.trigger_price
        size = self.check_trade_size(order.instrument, order.size)

        response = self.api.order.limit(
            accountID=self.ACCOUNT_ID,
            instrument=order.instrument,
            units=order.direction * size,
            price=str(price),
            takeProfitOnFill=take_profit_details,
            triggerCondition=trigger_condition,
            **stop_loss_order
        )
        return response

    def _modify_trade(self, order):
        """Modifies the take profit and/or stop loss of an existing trade.

        Parameters
        ----------
        order : TYPE
            DESCRIPTION.
        """
        # Get ID of trade to modify
        modify_trade_id = order.related_orders
        trade = self.api.trade.get(
            accountID=self.ACCOUNT_ID, tradeSpecifier=modify_trade_id
        ).body["trade"]

        if order.take_profit is not None:
            # Modify trade take-profit
            tpID = trade.takeProfitOrder.id

            # Cancel existing TP
            self.api.order.cancel(self.ACCOUNT_ID, tpID)

            # Create new TP
            tp_price = self._check_precision(order.instrument, order.take_profit)
            new_tp_order = self.api.order.TakeProfitOrder(
                tradeID=str(modify_trade_id), price=str(tp_price)
            )
            response = self.api.order.create(
                accountID=self.ACCOUNT_ID, order=new_tp_order
            )
            self._check_response(response)

        if order.stop_loss is not None:
            # Modify trade stop-loss
            slID = trade.stopLossOrder.id

            # Cancel existing SL
            self.api.order.cancel(self.ACCOUNT_ID, slID)

            # Create new SL
            sl_price = self._check_precision(order.instrument, order.stop_loss)
            new_sl_order = self.api.order.StopLossOrder(
                tradeID=str(modify_trade_id), price=str(sl_price)
            )
            response = self.api.order.create(
                accountID=self.ACCOUNT_ID, order=new_sl_order
            )
            self._check_response(response)

    def _get_stop_loss_order(self, order: Order) -> dict:
        """Constructs stop loss order dictionary."""
        self._check_connection()
        if order.stop_type is not None:
            price = self._check_precision(order.instrument, order.stop_loss)

            if order.stop_type == "trailing":
                # Trailing stop loss order
                SL_type = "trailingStopLossOnFill"

                # Calculate stop loss distance
                if order.stop_distance is None:
                    # Calculate stop distance from stop loss price provided
                    if order._working_price is not None:
                        working_price = order._working_price
                    else:
                        if order.order_type == "market":
                            # Get current market price
                            last = self._get_price(order.instrument)
                            working_price = (
                                last.closeoutBid
                                if order.direction < 0
                                else last.closeoutAsk
                            )
                        elif order.order_type in ["limit", "stop-limit"]:
                            working_price = order.order_limit_price
                    distance = abs(working_price - order.stop_loss)

                else:
                    # Calculate distance using provided pip distance
                    pip_value = 10 ** self.get_pip_location(order.instrument)
                    distance = abs(order.stop_distance * pip_value)

                # Construct stop loss order details
                distance = self._check_precision(order.instrument, distance)
                SL_details = {"distance": str(distance), "type": "TRAILING_STOP_LOSS"}
            else:
                SL_type = "stopLossOnFill"
                SL_details = {"price": str(price)}

            stop_loss_order = {SL_type: SL_details}

        else:
            stop_loss_order = {}

        return stop_loss_order

    def _get_take_profit_details(self, order: Order) -> dict:
        """Constructs take profit details dictionary."""
        self._check_connection()
        if order.take_profit is not None:
            price = self._check_precision(order.instrument, order.take_profit)
            take_profit_details = {"price": str(price)}
        else:
            take_profit_details = None

        return take_profit_details

    def _check_response(self, response):
        """Checks API response (currently only for placing orders)."""
        if response.status != 201:
            message = response.body["errorMessage"]
        else:
            message = "Success."

        output = {"Status": response.status, "Message": message}
        # TODO - print errors
        return output

    def _close_position(self, instrument, long_units=None, short_units=None, **kwargs):
        """Closes all open positions on an instrument."""
        self._check_connection()
        # Check if the position is long or short
        # Temp code to close all positions
        # Close all long units
        response = self.api.position.close(
            accountID=self.ACCOUNT_ID, instrument=instrument, longUnits="ALL"
        )

        # Close all short units
        response = self.api.position.close(
            accountID=self.ACCOUNT_ID, instrument=instrument, shortUnits="ALL"
        )

        # TODO - the code below makes no sense currently; specifically,
        # position.long.Units ????

        # open_position = self.get_open_positions(instrument)

        # if len(open_position) > 0:
        #     position = open_position['position']

        #     if long_units is None:
        #         long_units  = position.long.units
        #     if short_units is None:
        #         short_units = position.short.units

        #     if long_units > 0:
        #         response = self.api.position.close(accountID=self.ACCOUNT_ID,
        #                                            instrument=instrument,
        #                                            longUnits="ALL")

        #     elif short_units > 0:
        #         response = self.api.position.close(accountID=self.ACCOUNT_ID,
        #                                            instrument=instrument,
        #                                            shortUnits="ALL")

        #     else:
        #         print("There is no current position with {} to close.".format(instrument))
        #         response = None
        # else:
        #     response = None

        return response

    def _get_precision(self, instrument: str):
        """Returns the allowable precision for a given pair."""
        self._check_connection()
        response = self.api.account.instruments(
            accountID=self.ACCOUNT_ID, instruments=instrument
        )
        precision = response.body["instruments"][0].displayPrecision
        return precision

    def _check_precision(self, instrument, price):
        """Modify a price based on required ordering precision for pair."""
        N = self._get_precision(instrument)
        corrected_price = round(price, N)
        return corrected_price

    def _get_order_book(self, instrument: str):
        """Returns the order book of the instrument specified."""
        orderbook = self.autodata.L2(instrument=instrument)
        return orderbook

    def _get_position_book(self, instrument: str):
        """Returns the position book of the instrument specified."""
        response = self.api.instrument.position_book(instrument)
        return response.body["positionBook"]

    def _get_price(self, instrument: str):
        """Returns the current price of the instrument."""
        response = self.api.pricing.get(
            accountID=self.ACCOUNT_ID, instruments=instrument
        )
        return response.body["prices"][0]



import numpy as np
import pandas as pd
from finta import TA
from typing import Union


def supertrend(
    data: pd.DataFrame,
    period: int = 10,
    ATR_multiplier: float = 3.0,
    source: pd.Series = None,
) -> pd.DataFrame:
    """SuperTrend indicator, ported from the SuperTrend indicator by
    KivancOzbilgic on TradingView.

    Parameters
    ----------
    data : pd.DataFrame
        The OHLC data.
    period : int, optional
        The lookback period. The default is 10.
    ATR_multiplier : int, optional
        The ATR multiplier. The default is 3.0.
    source : pd.Series, optional
        The source series to use in calculations. If None, hl/2 will be
        used. The default is None.

    Returns
    -------
    supertrend_df : pd.DataFrame
        A Pandas DataFrame of containing the SuperTrend indicator, with
        columns of 'uptrend' and 'downtrend' containing uptrend/downtrend
        support/resistance levels, and 'trend', containing -1/1 to indicate
        the current implied trend.

    References
    ----------
    https://www.tradingview.com/script/r6dAP7yi/
    """

    if source is None:
        source = (data.High.values + data.Low.values) / 2

    # Calculate ATR
    atr = TA.ATR(data, period)

    up = source - (ATR_multiplier * atr)
    up_list = [up[0]]
    up_times = [data.index[0]]
    N_up = 0

    dn = source + (ATR_multiplier * atr)
    dn_list = [dn[0]]
    dn_times = [data.index[0]]
    N_dn = 0

    trend = 1
    trend_list = [trend]

    for i in range(1, len(data)):
        if trend == 1:
            if data.Close.values[i] > max(up[N_up:i]):
                up_list.append(max(up[N_up:i]))
                up_times.append(data.index[i])

                dn_list.append(np.nan)
                dn_times.append(data.index[i])
            else:
                trend = -1
                N_dn = i
                dn_list.append(dn[i])
                dn_times.append(data.index[i])

                up_list.append(np.nan)
                up_times.append(data.index[i])

        else:
            if data.Close.values[i] < min(dn[N_dn:i]):
                dn_list.append(min(dn[N_dn:i]))
                dn_times.append(data.index[i])

                up_list.append(np.nan)
                up_times.append(data.index[i])
            else:
                trend = 1
                N_up = i
                up_list.append(up[i])
                up_times.append(data.index[i])

                dn_list.append(np.nan)
                dn_times.append(data.index[i])

        trend_list.append(trend)

    supertrend_df = pd.DataFrame(
        {"uptrend": up_list, "downtrend": dn_list, "trend": trend_list}, index=up_times
    )
    return supertrend_df


def halftrend(
    data: pd.DataFrame, amplitude: int = 2, channel_deviation: float = 2
) -> pd.DataFrame:
    """HalfTrend indicator, ported from the HalfTrend indicator by
    Alex Orekhov (everget) on TradingView.

    Parameters
    ----------
    data : pd.DataFrame
        OHLC price data.
    amplitude : int, optional
        The lookback window. The default is 2.
    channel_deviation : float, optional
        The ATR channel deviation factor. The default is 2.

    Returns
    -------
    htdf : TYPE
        DESCRIPTION.

    References
    ----------
    https://www.tradingview.com/script/U1SJ8ubc-HalfTrend/
    """

    # Initialisation
    atr2 = TA.ATR(data, 100) / 2
    dev = channel_deviation * atr2
    high_price = data.High.rolling(amplitude).max().fillna(0)
    low_price = data.Low.rolling(amplitude).min().fillna(0)
    highma = TA.SMA(data, period=amplitude, column="High")
    lowma = TA.SMA(data, period=amplitude, column="Low")

    trend = np.zeros(len(data))
    next_trend = np.zeros(len(data))
    max_low_price = np.zeros(len(data))
    max_low_price[0] = data.Low[0]
    min_high_price = np.zeros(len(data))
    min_high_price[0] = data.High[0]

    for i in range(1, len(data)):
        if next_trend[i - 1] == 1:
            max_low_price[i] = max(low_price[i - 1], max_low_price[i - 1])

            if highma[i] < max_low_price[i] and data.Close[i] < data.Low[i - 1]:
                trend[i] = 1
                next_trend[i] = 0
                min_high_price[i] = high_price[i]
            else:
                # assign previous values again
                trend[i] = trend[i - 1]
                next_trend[i] = next_trend[i - 1]
                min_high_price[i] = min_high_price[i - 1]
        else:
            min_high_price[i] = min(high_price[i - 1], min_high_price[i - 1])

            if lowma[i] > min_high_price[i] and data.Close[i] > data.High[i - 1]:
                trend[i] = 0
                next_trend[i] = 1
                max_low_price[i] = low_price[i]
            else:
                # assign previous values again
                trend[i] = trend[i - 1]
                next_trend[i] = next_trend[i - 1]
                max_low_price[i] = max_low_price[i - 1]

    up = np.zeros(len(data))
    up[0] = max_low_price[0]
    down = np.zeros(len(data))
    down[0] = min_high_price[0]
    atr_high = np.zeros(len(data))
    atr_low = np.zeros(len(data))

    for i in range(1, len(data)):
        if trend[i] == 0:
            if trend[i - 1] != 0:
                up[i] = down[i - 1]
            else:
                up[i] = max(max_low_price[i - 1], up[i - 1])

            atr_high[i] = up[i] + dev[i]
            atr_low[i] = up[i] - dev[i]

        else:
            if trend[i - 1] != 1:
                down[i] = up[i - 1]
            else:
                down[i] = min(min_high_price[i - 1], down[i - 1])

            atr_high[i] = down[i] + dev[i]
            atr_low[i] = down[i] - dev[i]

    halftrend = np.where(trend == 0, up, down)
    buy = np.where((trend == 0) & (np.roll(trend, 1) == 1), 1, 0)
    sell = np.where((trend == 1) & (np.roll(trend, 1) == 0), 1, 0)

    # Construct DataFrame
    htdf = pd.DataFrame(
        data={
            "halftrend": halftrend,
            "atrHigh": np.nan_to_num(atr_high),
            "atrLow": np.nan_to_num(atr_low),
            "buy": buy,
            "sell": sell,
        },
        index=data.index,
    )

    # Clear false leading signals
    htdf.buy.values[:100] = np.zeros(100)
    htdf.sell.values[:100] = np.zeros(100)

    # Replace leading zeroes with nan
    htdf["atrHigh"] = htdf.atrHigh.replace(to_replace=0, value=float("nan"))
    htdf["atrLow"] = htdf.atrLow.replace(to_replace=0, value=float("nan"))

    return htdf


def range_filter(
    data: pd.DataFrame,
    range_qty: float = 2.618,
    range_period: int = 14,
    smooth_range: bool = True,
    smooth_period: int = 27,
    av_vals: bool = False,
    av_samples: int = 2,
    mov_source: str = "body",
    filter_type: int = 1,
) -> pd.DataFrame:
    """Price range filter, ported from the Range Filter [DW] indicator by
    DonovanWall on TradingView. The indicator was designed to filter out
    minor price action for a clearer view of trends.

    Parameters
    ----------
    data : pd.DataFrame
        The OHLC price data.
    range_qty : float, optional
        The range size. The default is 2.618.
    range_period : int, optional
        The range period. The default is 14.
    smooth_range : bool, optional
        Smooth the price range. The default is True.
    smooth_period : int, optional
        The smooting period. The default is 27.
    av_vals : bool, optional
        Average values. The default is False.
    av_samples : int, optional
        The number of average samples to use. The default is 2.
    mov_source : str, optional
        The price movement source ('body' or 'wicks'). The default is 'body'.
    filter_type : int, optional
        The filter type to use in calculations (1 or 2). The default is 1.

    Returns
    -------
    rfi : pd.DataFrame
        A dataframe containing the range filter indicator bounds.

    References
    ----------
    https://www.tradingview.com/script/lut7sBgG-Range-Filter-DW/
    """
    # Get high and low values
    if mov_source == "body":
        high_val = data.Close
        low_val = data.Close
    elif mov_source == "wicks":
        high_val = data.High
        low_val = data.Low

    # Get filter values
    rng = _range_size(
        (high_val + low_val) / 2, "AverageChange", range_qty, range_period
    )
    rfi = _calculate_range_filter(
        high_val,
        low_val,
        rng,
        range_period,
        filter_type,
        smooth_range,
        smooth_period,
        av_vals,
        av_samples,
    )

    return rfi


def bullish_engulfing(data: pd.DataFrame, detection: str = None):
    """Bullish engulfing pattern detection."""
    if detection == "SMA50":
        sma50 = sma(data.Close.values, 50)
        down_trend = np.where(data.Close.values < sma50, True, False)

    elif detection == "SMA50/200":
        sma50 = sma(data.Close.values, 50)
        sma200 = sma(data.Close.values, 200)

        down_trend = np.where(
            (data.Close.values < sma50) & (data.Close.values < sma200), True, False
        )
    else:
        down_trend = np.full(len(data), True)

    body_len = 14  # ema depth for bodyAvg

    body_high = np.maximum(data.Close.values, data.Open.values)
    body_low = np.minimum(data.Close.values, data.Open.values)
    body = body_high - body_low

    body_avg = ema(body, body_len)
    short_body = body < body_avg
    long_body = body > body_avg

    white_body = data.Open.values < data.Close.values
    black_body = data.Open.values > data.Close.values

    inside_bar = [False]
    for i in range(1, len(data)):
        val = (body_high[i - 1] > body_high[i]) and (body_low[i - 1] < body_low[i])
        inside_bar.append(val)

    engulfing_bullish = [False]
    for i in range(1, len(data)):
        condition = (
            down_trend[i]
            & white_body[i]
            & long_body[i]
            & black_body[i - 1]
            & short_body[i - 1]
            & (data.Close.values[i] >= data.Open.values[i - 1])
            & (data.Open.values[i] <= data.Close.values[i - 1])
            & (
                (data.Close.values[i] > data.Open.values[i - 1])
                | (data.Open.values[i] < data.Close.values[i - 1])
            )
        )

        engulfing_bullish.append(condition)

    return engulfing_bullish


def bearish_engulfing(data: pd.DataFrame, detection: str = None):
    """Bearish engulfing pattern detection."""
    if detection == "SMA50":
        sma50 = sma(data.Close.values, 50)
        up_trend = np.where(data.Close.values > sma50, True, False)
    elif detection == "SMA50/200":
        sma50 = sma(data.Close.values, 50)
        sma200 = sma(data.Close.values, 200)

        up_trend = np.where(
            (data.Close.values > sma50) & (data.Close.values > sma200), True, False
        )
    else:
        up_trend = np.full(len(data), True)

    body_len = 14  # ema depth for bodyAvg
    body_high = np.maximum(data.Close.values, data.Open.values)
    body_low = np.minimum(data.Close.values, data.Open.values)
    body = body_high - body_low

    body_avg = ema(body, body_len)
    short_body = body < body_avg
    long_body = body > body_avg

    white_body = data.Open.values < data.Close.values
    black_body = data.Open.values > data.Close.values

    inside_bar = [False]
    for i in range(1, len(data)):
        val = (body_high[i - 1] > body_high[i]) and (body_low[i - 1] < body_low[i])
        inside_bar.append(val)

    engulfing_bearish = [False]
    for i in range(1, len(data)):
        condition = (
            up_trend[i]
            & black_body[i]
            & long_body[i]
            & white_body[i - 1]
            & short_body[i - 1]
            & (data.Close.values[i] <= data.Open.values[i - 1])
            & (data.Open.values[i] >= data.Close.values[i - 1])
            & (
                (data.Close.values[i] < data.Open.values[i - 1])
                | (data.Open.values[i] > data.Close.values[i - 1])
            )
        )

        engulfing_bearish.append(condition)

    return engulfing_bearish


def find_swings(data: pd.DataFrame, n: int = 2) -> pd.DataFrame:
    """Locates swings in the inputted data using a moving average gradient
    method.

    Parameters
    ----------
    data : pd.DataFrame | pd.Series | list | np.array
        An OHLC dataframe of price, or an array/list/Series of data from an
        indicator (eg. RSI).
    n : int, optional
        The moving average period. The default is 2.

    Returns
    -------
    swing_df : pd.DataFrame
        A dataframe containing the swing levels detected.
    """
    # Prepare data
    if isinstance(data, pd.DataFrame):
        # OHLC data
        hl2 = (data.High.values + data.Low.values) / 2
        swing_data = pd.Series(ema(hl2, n), index=data.index)
        low_data = data.Low.values
        high_data = data.High.values

    elif isinstance(data, pd.Series):
        # Pandas series data
        swing_data = pd.Series(ema(data, n), index=data.index)
        low_data = data
        high_data = data

    else:
        # Find swings in alternative data source
        data = pd.Series(data)

        # Define swing data
        swing_data = pd.Series(ema(data, n), index=data.index)
        low_data = data
        high_data = data

    signed_grad = np.sign((swing_data - swing_data.shift(1)).fillna(method="bfill"))
    swings = (signed_grad != signed_grad.shift(1).fillna(method="bfill")) * -signed_grad

    # Calculate swing extrema
    lows = []
    highs = []
    for i, swing in enumerate(swings):
        if swing < 0:
            # Down swing, find low price
            highs.append(0)
            lows.append(min(low_data[i - n : i]))
        elif swing > 0:
            # Up swing, find high price
            highs.append(max(high_data[i - n : i]))
            lows.append(0)
        else:
            # Price movement
            highs.append(0)
            lows.append(0)

    # Determine last swing
    trend = rolling_signal_list(-swings)
    swings_list = merge_signals(lows, highs)
    last_swing = rolling_signal_list(swings_list)

    # Need to return both a last swing low and last swing high list
    last_low = rolling_signal_list(lows)
    last_high = rolling_signal_list(highs)

    swing_df = pd.DataFrame(
        data={"Highs": last_high, "Lows": last_low, "Last": last_swing, "Trend": trend},
        index=swing_data.index,
    )

    return swing_df


def classify_swings(swing_df: pd.DataFrame, tol: int = 0) -> pd.DataFrame:
    """Classifies a dataframe of swings (from find_swings) into higher-highs,
    lower-highs, higher-lows and lower-lows.


    Parameters
    ----------
    swing_df : pd.DataFrame
        The dataframe returned by find_swings.
    tol : int, optional
        The classification tolerance. The default is 0.

    Returns
    -------
    swing_df : pd.DataFrame
        A dataframe containing the classified swings.
    """
    # Create copy of swing dataframe
    swing_df = swing_df.copy()

    new_level = np.where(swing_df.Last != swing_df.Last.shift(), 1, 0)

    candles_since_last = candles_between_crosses(new_level)

    # Add column 'candles since last swing' CSLS
    swing_df["CSLS"] = candles_since_last

    # Find strong Support and Resistance zones
    swing_df["Support"] = (swing_df.CSLS > tol) & (swing_df.Trend == 1)
    swing_df["Resistance"] = (swing_df.CSLS > tol) & (swing_df.Trend == -1)

    # Find higher highs and lower lows
    swing_df["Strong_lows"] = (
        swing_df["Support"] * swing_df["Lows"]
    )  # Returns high values when there is a strong support
    swing_df["Strong_highs"] = (
        swing_df["Resistance"] * swing_df["Highs"]
    )  # Returns high values when there is a strong support

    # Remove duplicates to preserve indexes of new levels
    swing_df["FSL"] = unroll_signal_list(
        swing_df["Strong_lows"]
    )  # First of new strong lows
    swing_df["FSH"] = unroll_signal_list(
        swing_df["Strong_highs"]
    )  # First of new strong highs

    # Now compare each non-zero value to the previous non-zero value.
    low_change = np.sign(swing_df.FSL) * (
        swing_df.FSL
        - swing_df.Strong_lows.replace(to_replace=0, method="ffill").shift()
    )
    high_change = np.sign(swing_df.FSH) * (
        swing_df.FSH
        - swing_df.Strong_highs.replace(to_replace=0, method="ffill").shift()
    )

    swing_df["LL"] = np.where(low_change < 0, True, False)
    swing_df["HL"] = np.where(low_change > 0, True, False)
    swing_df["HH"] = np.where(high_change > 0, True, False)
    swing_df["LH"] = np.where(high_change < 0, True, False)

    return swing_df


def detect_divergence(
    classified_price_swings: pd.DataFrame,
    classified_indicator_swings: pd.DataFrame,
    tol: int = 2,
    method: int = 0,
) -> pd.DataFrame:
    """Detects divergence between price swings and swings in an indicator.

    Parameters
    ----------
    classified_price_swings : pd.DataFrame
        The output from classify_swings using OHLC data.
    classified_indicator_swings : pd.DataFrame
        The output from classify_swings using indicator data.
    tol : int, optional
        The number of candles which conditions must be met within. The
        default is 2.
    method : int, optional
        The method to use when detecting divergence (0 or 1). The default is 0.

    Raises
    ------
    Exception
        When an unrecognised method of divergence detection is requested.

    Returns
    -------
    divergence : pd.DataFrame
        A dataframe containing divergence signals.

    Notes
    -----
    Options for the method include:
        0: use both price and indicator swings to detect divergence (default)

        1: use only indicator swings to detect divergence (more responsive)
    """
    if method == 0:
        regular_bullish = []
        regular_bearish = []
        hidden_bullish = []
        hidden_bearish = []

        for i in range(len(classified_price_swings)):
            # Look backwards in each

            # REGULAR BULLISH DIVERGENCE
            if (
                sum(classified_price_swings["LL"][i - tol : i])
                + sum(classified_indicator_swings["HL"][i - tol : i])
                > 1
            ):
                regular_bullish.append(True)
            else:
                regular_bullish.append(False)

            # REGULAR BEARISH DIVERGENCE
            if (
                sum(classified_price_swings["HH"][i - tol : i])
                + sum(classified_indicator_swings["LH"][i - tol : i])
                > 1
            ):
                regular_bearish.append(True)
            else:
                regular_bearish.append(False)

            # HIDDEN BULLISH DIVERGENCE
            if (
                sum(classified_price_swings["HL"][i - tol : i])
                + sum(classified_indicator_swings["LL"][i - tol : i])
                > 1
            ):
                hidden_bullish.append(True)
            else:
                hidden_bullish.append(False)

            # HIDDEN BEARISH DIVERGENCE
            if (
                sum(classified_price_swings["LH"][i - tol : i])
                + sum(classified_indicator_swings["HH"][i - tol : i])
                > 1
            ):
                hidden_bearish.append(True)
            else:
                hidden_bearish.append(False)

        divergence = pd.DataFrame(
            data={
                "regularBull": unroll_signal_list(regular_bullish),
                "regularBear": unroll_signal_list(regular_bearish),
                "hiddenBull": unroll_signal_list(hidden_bullish),
                "hiddenBear": unroll_signal_list(hidden_bearish),
            },
            index=classified_price_swings.index,
        )
    elif method == 1:
        # Use indicator swings only to detect divergence
        for i in range(len(classified_price_swings)):

            price_at_indi_lows = (
                classified_indicator_swings["FSL"] != 0
            ) * classified_price_swings["Lows"]
            price_at_indi_highs = (
                classified_indicator_swings["FSH"] != 0
            ) * classified_price_swings["Highs"]

            # Determine change in price between indicator lows
            price_at_indi_lows_change = np.sign(price_at_indi_lows) * (
                price_at_indi_lows
                - price_at_indi_lows.replace(to_replace=0, method="ffill").shift()
            )
            price_at_indi_highs_change = np.sign(price_at_indi_highs) * (
                price_at_indi_highs
                - price_at_indi_highs.replace(to_replace=0, method="ffill").shift()
            )

            # DETECT DIVERGENCES
            regular_bullish = (classified_indicator_swings["HL"]) & (
                price_at_indi_lows_change < 0
            )
            regular_bearish = (classified_indicator_swings["LH"]) & (
                price_at_indi_highs_change > 0
            )
            hidden_bullish = (classified_indicator_swings["LL"]) & (
                price_at_indi_lows_change > 0
            )
            hidden_bearish = (classified_indicator_swings["HH"]) & (
                price_at_indi_highs_change < 0
            )

        divergence = pd.DataFrame(
            data={
                "regularBull": regular_bullish,
                "regularBear": regular_bearish,
                "hiddenBull": hidden_bullish,
                "hiddenBear": hidden_bearish,
            },
            index=classified_price_swings.index,
        )

    else:
        raise Exception("Error: unrecognised method of divergence detection.")

    return divergence


def autodetect_divergence(
    ohlc: pd.DataFrame,
    indicator_data: pd.DataFrame,
    tolerance: int = 1,
    method: int = 0,
) -> pd.DataFrame:
    """A wrapper method to automatically detect divergence from inputted OHLC price
    data and indicator data.

    Parameters
    ----------
    ohlc : pd.DataFrame
        A dataframe of OHLC price data.
    indicator_data : pd.DataFrame
        dataframe of indicator data.
    tolerance : int, optional
        A parameter to control the lookback when detecting divergence.
        The default is 1.
    method : int, optional
        The divergence detection method. Set to 0 to use both price and
        indicator swings to detect divergence. Set to 1 to use only indicator
        swings to detect divergence. The default is 0.

    Returns
    -------
    divergence : pd.DataFrame
        A DataFrame containing columns 'regularBull', 'regularBear',
        'hiddenBull' and 'hiddenBear'.

    See Also
    --------
    autotrader.indicators.find_swings
    autotrader.indicators.classify_swings
    autotrader.indicators.detect_divergence

    """

    # Price swings
    price_swings = find_swings(ohlc)
    price_swings_classified = classify_swings(price_swings)

    # Indicator swings
    indicator_swings = find_swings(indicator_data, data_type="other")
    indicator_classified = classify_swings(indicator_swings)

    # Detect divergence
    divergence = detect_divergence(
        price_swings_classified, indicator_classified, tol=tolerance, method=method
    )

    return divergence


def heikin_ashi(data: pd.DataFrame):
    """Calculates the Heikin-Ashi candlesticks from Japanese candlestick
    data.
    """
    # Create copy of data to prevent overwriting
    working_data = data.copy()

    # Calculate Heikin Ashi candlesticks
    ha_close = 0.25 * (
        working_data.Open + working_data.Low + working_data.High + working_data.Close
    )
    ha_open = 0.5 * (working_data.Open + working_data.Close)
    ha_high = np.maximum(
        working_data.High.values, working_data.Close.values, working_data.Open.values
    )
    ha_low = np.minimum(
        working_data.Low.values, working_data.Close.values, working_data.Open.values
    )

    ha_data = pd.DataFrame(
        data={"Open": ha_open, "High": ha_high, "Low": ha_low, "Close": ha_close},
        index=working_data.index,
    )

    return ha_data


def ha_candle_run(ha_data: pd.DataFrame):
    """Returns a list for the number of consecutive green and red
    Heikin-Ashi candles.

    Parameters
    ----------
    ha_data: pd.DataFrame
        The Heikin Ashi OHLC data.

    See Also
    --------
    heikin_ashi
    """
    green_candle = np.where(ha_data.Close - ha_data.Open > 0, 1, 0)
    red_candle = np.where(ha_data.Close - ha_data.Open < 0, 1, 0)

    green_run = []
    red_run = []

    green_sum = 0
    red_sum = 0

    for i in range(len(ha_data)):
        if green_candle[i] == 1:
            green_sum += 1
        else:
            green_sum = 0

        if red_candle[i] == 1:
            red_sum += 1
        else:
            red_sum = 0

        green_run.append(green_sum)
        red_run.append(red_sum)

    return green_run, red_run


def N_period_high(data: pd.DataFrame, N: int):
    """Returns the N-period high."""
    highs = data.High.rolling(N).max()
    return highs


def N_period_low(data: pd.DataFrame, N: int):
    """Returns the N-period low."""
    lows = data.Low.rolling(N).min()
    return lows


def crossover(ts1: pd.Series, ts2: pd.Series) -> pd.Series:
    """Locates where two timeseries crossover each other, returning 1 when
    list_1 crosses above list_2, and -1 for when list_1 crosses below list_2.

    Parameters
    ----------
    ts1 : pd.Series
        The first timeseries.
    ts2 : pd.Series
        The second timeseries.

    Returns
    -------
    crossovers : pd.Series
        The crossover series.
    """

    signs = np.sign(ts1 - ts2)
    crossovers = pd.Series(data=signs * (signs != signs.shift(1)), name="crossovers")

    return crossovers


def cross_values(
    ts1: Union[list, pd.Series],
    ts2: Union[list, pd.Series],
    ts_crossover: Union[list, pd.Series] = None,
) -> Union[list, pd.Series]:
    """Returns the approximate value of the point where the two series cross.

    Parameters
    ----------
    ts1 : list | pd.Series
        The first timeseries..
    ts2 : list | pd.Series
        The second timeseries..
    ts_crossover : list | pd.Series, optional
        The crossovers between timeseries 1 and timeseries 2.

    Returns
    -------
    cross_points : list | pd.Series
        The values at which crossovers occur.
    """

    if ts_crossover is None:
        ts_crossover = crossover(ts1, ts2)

    last_cross_point = ts1[0]
    cross_points = [last_cross_point]
    for i in range(1, len(ts_crossover)):
        if ts_crossover[i] != 0:
            i0 = 0
            m_a = ts1[i] - ts1[i - 1]
            m_b = ts2[i] - ts2[i - 1]
            ix = (ts2[i - 1] - ts1[i - 1]) / (m_a - m_b) + i0

            cross_point = m_a * (ix - i0) + ts1[i - 1]

            last_cross_point = cross_point

        else:
            cross_point = last_cross_point

        cross_points.append(cross_point)

    # Replace nans with 0
    cross_points = [0 if x != x else x for x in cross_points]

    if isinstance(ts1, pd.Series):
        # Convert to Series
        cross_points = pd.Series(data=cross_points, index=ts1.index, name="crossval")

    return cross_points


def candles_between_crosses(crosses: Union[list, pd.Series]) -> Union[list, pd.Series]:
    """Returns a rolling sum of candles since the last cross/signal occurred.

    Parameters
    ----------
    crosses : list | pd.Series
        The list or Series containing crossover signals.

    Returns
    -------
    counts : TYPE
        The rolling count of bars since the last crossover signal.

    See Also
    ---------
    indicators.crossover
    """

    count = 0
    counts = []

    for i in range(len(crosses)):
        if crosses[i] == 0:
            # Change in signal - reset count
            count += 1
        else:
            count = 0

        counts.append(count)

    if isinstance(crosses, pd.Series):
        # Convert to Series
        counts = pd.Series(data=counts, index=crosses.index, name="counts")

    return counts


def rolling_signal_list(signals: Union[list, pd.Series]) -> list:
    """Returns a list which repeats the previous signal, until a new
    signal is given.

    Parameters
    ----------
    signals : list | pd.Series
        A series of signals. Zero values are treated as 'no signal'.

    Returns
    -------
    list
        A list of rolled signals.

    Examples
    --------
    >>> rolling_signal_list([0,1,0,0,0,-1,0,0,1,0,0])
        [0, 1, 1, 1, 1, -1, -1, -1, 1, 1, 1]

    """
    rolling_signals = [0]
    last_signal = rolling_signals[0]

    for i in range(1, len(signals)):
        if signals[i] != 0:
            last_signal = signals[i]

        rolling_signals.append(last_signal)

    if isinstance(signals, pd.Series):
        rolling_signals = pd.Series(data=rolling_signals, index=signals.index)

    return rolling_signals


def unroll_signal_list(signals: Union[list, pd.Series]) -> np.array:
    """Unrolls a rolled signal list.

    Parameters
    ----------
    signals : Union[list, pd.Series]
        DESCRIPTION.

    Returns
    -------
    unrolled_signals : np.array
        The unrolled signal series.

    See Also
    --------
    This function is the inverse of rolling_signal_list.

    Examples
    --------
    >>> unroll_signal_list([0, 1, 1, 1, 1, -1, -1, -1, 1, 1, 1])
        array([ 0.,  1.,  0.,  0.,  0., -1.,  0.,  0.,  1.,  0.,  0.])

    """
    unrolled_signals = np.zeros(len(signals))
    for i in range(len(signals)):
        if signals[i] != signals[i - 1]:
            unrolled_signals[i] = signals[i]

    if isinstance(signals, pd.Series):
        unrolled_signals = pd.Series(data=unrolled_signals, index=signals.index)

    return unrolled_signals


def merge_signals(signal_1: list, signal_2: list) -> list:
    """Returns a single signal list which has merged two signal lists.

    Parameters
    ----------
    signal_1 : list
        The first signal list.
    signal_2 : list
        The second signal list.

    Returns
    -------
    merged_signal_list : list
        The merged result of the two inputted signal series.

    Examples
    --------
    >>> s1 = [1,0,0,0,1,0]
    >>> s2 = [0,0,-1,0,0,-1]
    >>> merge_signals(s1, s2)
        [1, 0, -1, 0, 1, -1]

    """
    merged_signal_list = signal_1.copy()
    for i in range(len(signal_1)):
        if signal_2[i] != 0:
            merged_signal_list[i] = signal_2[i]

    return merged_signal_list


def build_grid_price_levels(
    grid_origin: float,
    grid_space: float,
    grid_levels: int,
    grid_price_space: float = None,
    pip_value: float = 0.0001,
) -> np.array:
    """Generates grid price levels."""
    # Calculate grid spacing in price units
    if grid_price_space is None:
        grid_price_space = grid_space * pip_value

    # Generate order_limit_price list
    grid_price_levels = np.linspace(
        grid_origin - grid_levels * grid_price_space,
        grid_origin + grid_levels * grid_price_space,
        2 * grid_levels + 1,
    )

    return grid_price_levels


def build_grid(
    grid_origin: float,
    grid_space: float,
    grid_levels: int,
    order_direction: int,
    order_type: str = "stop-limit",
    grid_price_space: float = None,
    pip_value: float = 0.0001,
    take_distance: float = None,
    stop_distance: float = None,
    stop_type: str = None,
) -> dict:
    """Generates a grid of orders.

    Parameters
    ----------
    grid_origin : float
        The origin of grid, specified as a price.
    grid_space : float
        The spacing between grid levels, specified as pip distance.
    grid_levels : int
        The number of grid levels either side of origin.
    order_direction : int
        The direction of each grid level order (1 for long, -1 for short).
    order_type : str, optional
        The order type of each grid level order. The default is 'stop-limit'.
    grid_price_space : float, optional
        The spacing between grid levels, specified as price units distance.
        The default is None.
    pip_value : float, optional
        The instrument-specific pip value. The default is 0.0001.
    take_distance : float, optional
        The distance (in pips) of each order's take profit. The default is None.
    stop_distance : float, optional
        The distance (in pips) of each order's stop loss. The default is None.
    stop_type : str, optional
        The stop loss type. The default is None.

    Returns
    -------
    grid : dict
        A dictionary containing all orders on the grid.

    """

    # Check if stop_distance was provided without a stop_type
    if stop_distance is not None and stop_type is None:
        # set stop_type to 'limit' by default
        stop_type = "limit"

    # Calculate grid spacing in price units
    if grid_price_space is None:
        grid_price_space = grid_space * pip_value

    # Generate order_limit_price list
    order_limit_prices = np.linspace(
        grid_origin - grid_levels * grid_price_space,
        grid_origin + grid_levels * grid_price_space,
        2 * grid_levels + 1,
    )

    # Construct nominal order
    nominal_order = {}
    nominal_order["order_type"] = order_type
    nominal_order["direction"] = order_direction
    nominal_order["stop_distance"] = stop_distance
    nominal_order["stop_type"] = stop_type
    nominal_order["take_distance"] = take_distance

    # Build grid
    grid = {}

    for order, limit_price in enumerate(order_limit_prices):
        grid[order] = nominal_order.copy()
        grid[order]["order_stop_price"] = order_limit_prices[order]
        grid[order]["order_limit_price"] = order_limit_prices[order]

    return grid


def merge_grid_orders(grid_1: np.array, grid_2: np.array) -> np.array:
    """Merges grid dictionaries into one and re-labels order numbers so each
    order number is unique.
    """
    order_offset = len(grid_1)
    grid = grid_1.copy()

    for order_no in grid_2:
        grid[order_no + order_offset] = grid_2[order_no]

    return grid


def last_level_crossed(data: pd.DataFrame, base: float) -> list:
    """Returns a list containing the last grid level touched.
    The grid levels are determined by the base input variable,
    which corresponds to the pip_space x pip_value.
    """
    last_level_crossed = np.nan
    levels_crossed = []
    for i in range(len(data)):
        high = data.High.values[i]
        low = data.Low.values[i]

        upper_prices = []
        lower_prices = []

        for price in [high, low]:
            upper_prices.append(base * np.ceil(price / base))
            lower_prices.append(base * np.floor(price / base))

        if lower_prices[0] != lower_prices[1]:
            # Candle has crossed a level
            last_level_crossed = lower_prices[0]

        levels_crossed.append(last_level_crossed)

    return levels_crossed


def build_multiplier_grid(
    origin: float,
    direction: int,
    multiplier: float,
    no_levels: int,
    precision: int,
    spacing: float,
) -> list:
    """Constructs grid levels with a multiplying grid space.

    Parameters
    ----------
    origin : float
        The origin of grid as price amount.

    direction : int
        The direction of grid (1 for long, -1 for short).

    multiplier : float
        The grid space multiplier when price moves away from the origin
        opposite to direction.

    no_levels : int
        The number of levels to calculate either side of the origin.

    precision : int
        The instrument precision (eg. 4 for most currencies, 2 for JPY).

    spacing : float
        The spacing of the grid in price units.
    """

    levels = [i for i in range(1, no_levels + 1)]

    pos_levels = [round(origin + direction * spacing * i, precision) for i in levels]
    neg_spaces = [spacing * multiplier ** (i) for i in levels]
    neg_levels = []
    prev_neg_level = origin
    for i in range(len(levels)):
        next_neg_level = prev_neg_level - direction * neg_spaces[i]
        prev_neg_level = next_neg_level
        neg_levels.append(round(next_neg_level, precision))

    grid = neg_levels + [origin] + pos_levels
    grid.sort()

    return grid


def last_level_touched(data: pd.DataFrame, grid: np.array) -> np.array:
    """Calculates the grid levels touched by price data."""
    # initialise with nan
    last_level_crossed = np.nan

    levels_touched = []
    for i in range(len(data)):
        high = data.High.values[i]
        low = data.Low.values[i]

        upper_prices = []
        lower_prices = []

        for price in [high, low]:
            # Calculate level above
            upper_prices.append(
                grid[next(x[0] for x in enumerate(grid) if x[1] > price)]
            )

            # calculate level below
            first_level_below_index = next(
                x[0] for x in enumerate(grid[::-1]) if x[1] < price
            )
            lower_prices.append(grid[-(first_level_below_index + 1)])

        if lower_prices[0] != lower_prices[1]:
            # Candle has crossed a level, since the level below the candle high
            # is different to the level below the candle low.
            # This essentially means the grid level is between candle low and high.
            last_level_crossed = lower_prices[0]

        levels_touched.append(last_level_crossed)

    return levels_touched


def stoch_rsi(
    data: pd.DataFrame,
    K_period: int = 3,
    D_period: int = 3,
    RSI_length: int = 14,
    Stochastic_length: int = 14,
):
    """Stochastic RSI indicator."""
    rsi1 = TA.RSI(data, period=RSI_length)
    stoch = stochastic(rsi1, rsi1, rsi1, Stochastic_length)

    K = sma(stoch, K_period)
    D = sma(K, D_period)

    return K, D


def stochastic(
    high: pd.Series, low: pd.Series, close: pd.Series, period: int = 14
) -> pd.Series:
    """Stochastics indicator."""
    K = np.zeros(len(high))

    for i in range(period, len(high)):
        low_val = min(low[i - period + 1 : i + 1])
        high_val = max(high[i - period + 1 : i + 1])

        K[i] = 100 * (close[i] - low_val) / (high_val - low_val)

    return K


def sma(data: pd.DataFrame, period: int = 14) -> list:
    """Smoothed Moving Average."""
    sma_list = []
    for i in range(len(data)):
        average = sum(data[i - period + 1 : i + 1]) / period
        sma_list.append(average)
    return sma_list


def ema(data: pd.DataFrame, period: int = 14, smoothing: int = 2) -> list:
    """Exponential Moving Average."""
    ema = [sum(data[:period]) / period]
    for price in data[period:]:
        ema.append(
            (price * (smoothing / (1 + period)))
            + ema[-1] * (1 - (smoothing / (1 + period)))
        )
    for i in range(period - 1):
        ema.insert(0, np.nan)
    return ema


def true_range(data: pd.DataFrame, period: int = 14):
    """True range."""
    high_low = data["High"] - data["Low"]
    high_close = np.abs(data["High"] - data["Close"].shift())
    low_close = np.abs(data["Low"] - data["Close"].shift())
    ranges = pd.concat([high_low, high_close, low_close], axis=1)
    true_range = np.max(ranges, axis=1)
    return true_range


def atr(data: pd.DataFrame, period: int = 14):
    """Average True Range."""
    tr = true_range(data, period)
    atr = tr.rolling(period).sum() / period
    return atr


def create_bricks(data: pd.DataFrame, brick_size: float = 0.002, column: str = "Close"):
    """Creates a dataframe of price-sized bricks.

    Parameters
    ----------
    data : pd.DataFrame
        The OHLC price data.
    brick_size : float, optional
        The brick size in price units. The default is 0.0020.
    column : str, optional
        The column of the OHLC to use. The default is 'Close'.

    Returns
    -------
    bricks : pd.DataFrame
        The Open and Close prices of each brick, indexed by brick close time.

    """
    brick_open = data[column][0]
    opens = [brick_open]
    close_times = [data.index[0]]
    for i in range(len(data)):
        price = data[column][i]
        price_diff = price - brick_open
        if abs(price_diff) > brick_size:
            # New brick(s)
            no_new_bricks = abs(int(price_diff / brick_size))
            for b in range(no_new_bricks):
                brick_close = brick_open + np.sign(price_diff) * brick_size
                brick_open = brick_close
                opens.append(brick_open)
                close_times.append(data.index[i])

    bricks = pd.DataFrame(data={"Open": opens, "Close": opens}, index=close_times)
    bricks["Close"] = bricks["Close"].shift(-1)

    return bricks


def _conditional_ema(x, condition=1, n=14, s=2):
    "Conditional sampling EMA functtion"
    if type(condition) == int:
        condition = condition * np.ones(len(x))

    ema = np.zeros(len(x))
    for i in range(1, len(x)):
        if condition[i]:
            ema[i] = (x[i] - ema[i - 1]) * (s / (1 + n)) + ema[i - 1]
        else:
            ema[i] = ema[i - 1]

    return pd.Series(ema, x.index, name=f"{n} period conditional EMA")


def _conditional_sma(x, condition=1, n=14):
    "Conditional sampling SMA functtion"

    if type(condition) == int:
        condition = condition * np.ones(len(x))

    # Calculate SMA
    sma = x.rolling(n).mean()

    # Filter by condition
    sma = sma * condition

    return sma


def _stdev(x, n):
    "Standard deviation function"
    sd = np.sqrt(_conditional_sma(x**2, 1, n) - _conditional_sma(x, 1, n) ** 2)
    return sd


def _range_size(x, scale="AverageChange", qty=2.618, n=14):
    "Range size function"

    if scale == "AverageChange":
        AC = _conditional_ema(abs(x - x.shift(1)), 1, n)
        rng_size = qty * AC
    elif scale == "ATR":
        tr = TA.TR(x)
        atr = _conditional_ema(tr, 1, n)
        rng_size = qty * atr
    elif scale == "StandardDeviation":
        sd = _stdev(x, n)
        rng_size = qty * sd

    return rng_size


def _calculate_range_filter(h, l, rng, n, rng_type, smooth, sn, av_rf, av_n):
    """Two type range filter function."""

    smoothed_range = _conditional_ema(rng, 1, sn)
    r = smoothed_range if smooth else rng
    r_filt = (h + l) / 2

    if rng_type == 1:
        for i in range(1, len(h)):
            if h[i] - r[i] > r_filt[i - 1]:
                r_filt[i] = h[i] - r[i]
            elif l[i] + r[i] < r_filt[i - 1]:
                r_filt[i] = l[i] + r[i]
            else:
                r_filt[i] = r_filt[i - 1]

    elif rng_type == 2:
        for i in range(1, len(h)):
            if h[i] >= r_filt[i - 1] + r[i]:
                r_filt[i] = (
                    r_filt[i - 1] + np.floor(abs(h[i] - r_filt[i - 1]) / r[i]) * r[i]
                )
            elif l[i] <= r_filt[i - 1] - r[i]:
                r_filt[i] = (
                    r_filt[i - 1] - np.floor(abs(l[i] - r_filt[i - 1]) / r[i]) * r[i]
                )
            else:
                r_filt[i] = r_filt[i - 1]

    # Define nominal values
    r_filt1 = r_filt.copy()
    hi_band1 = r_filt1 + r
    lo_band1 = r_filt1 - r

    # Calculate indicator for averaged filter changes
    r_filt2 = _conditional_ema(r_filt1, r_filt1 != r_filt1.shift(1), av_n)
    hi_band2 = _conditional_ema(hi_band1, r_filt1 != r_filt1.shift(1), av_n)
    lo_band2 = _conditional_ema(lo_band1, r_filt1 != r_filt1.shift(1), av_n)

    # Assign indicator
    rng_filt = r_filt2 if av_rf else r_filt1
    hi_band = hi_band2 if av_rf else hi_band1
    lo_band = lo_band2 if av_rf else lo_band1

    # Construct output
    rfi = pd.DataFrame(
        data={"upper": hi_band, "lower": lo_band, "rf": rng_filt}, index=rng_filt.index
    )

    # Classify filter direction
    rfi["fdir"] = np.sign(rfi.rf - rfi.rf.shift(1)).fillna(0)

    return rfi



from .autoplot import AutoPlot
from .autodata import AutoData
from .autotrader import AutoTrader
from .utilities import TradeAnalysis, DataStream
from .brokers.trading import (
    Order,
    IsolatedPosition,
    Position,
    LimitOrder,
    MarketOrder,
    StopLimitOrder,
    Trade,
)

# Broker imports
# CCXT
try:
    from .brokers.ccxt.broker import Broker as CCXTBroker
except:
    pass

# dYdX
try:
    from .brokers.dydx.broker import Broker as dYdXBroker
except:
    pass

# Interactive Brokers
try:
    from .brokers.ib.broker import Broker as IBroker
except:
    pass

# Oanda
try:
    from .brokers.oanda.broker import Broker as OandaBroker
except:
    pass

# Virtual broker
from .brokers.virtual.broker import Broker as VirtualBroker

# Define version number
__version__ = "0.9.1"



from autotrader.brokers.broker_utils import BrokerUtils


class Utils(BrokerUtils):
    def __init__(self, **kwargs):
        pass

    def __repr__(self):
        return "AutoTrader Virtual Broker Utilities"

    def __str__(self):
        return "AutoTrader Virtual Broker Utilities"

    def get_precision(self, instrument, *arg, **kwargs):
        """Returns the precision of the specified instrument."""
        unified_response = {"size": 2, "price": 5}
        return unified_response



from __future__ import annotations
import numpy as np
from datetime import datetime
from autotrader.brokers.broker_utils import BrokerUtils


class Order:
    """AutoTrader Order object.

    Attributes
    ----------
    instrument : str
        The trading instrument of the order.
    direction : int
        The direction of the order (1 for long, -1 for short).
    order_type : str
        The type of order. The default is 'market'.
    size : float
        The number of units.
    base_size : float
         The number of units, in the base currency (pre-HCF conversion).
    target_value : float
        The target value of the resulting trade, specified in the home
        currency of the account.
    order_limit_price : float
        The limit price of the order (for 'limit' and 'stop-limit' order
        types).
    order_stop_price : float
        The stop price of the order (for 'stop-limit' order types).
    order_price : float
        The price of the instrument when the order was placed.
    order_time : datetime
        The time at which the order was placed.
    stop_loss : float
        The price to set the stop-loss at.
    stop_distance : float
        The pip distance between the order price and the stop-loss.
    stop_type : str
        The type of stop-loss (limit or trailing). The default is 'limit'.
    take_profit : float
        The price to set the take-profit at.
    take_distance : float
        The pip distance between the order price and the take-profit.
    related_orders : list
        A list of related order/trade ID's.
    id : int
        The order ID.
    pip_value : float, optional
        The pip value of the product being traded. Specify this for non-FX
        products when using stop_distance/take_distance arguments. The default
        is None.
    currency : str
        The base currency of the order (IB only).
    secType : str
        The security type (IB only).
    contract_month : str
        The contract month string (IB only).
    localSymbol : str
        The exchange-specific instrument symbol (IB only).
    post_only : bool, optional
        Enforce that the order is placed as a maker order (dYdX only). The
        default is False.
    limit_fee : str, optional
        The maximum fee to accept as a percentage (dYdX only). The default
        is '0.015'.
    exchange : str
        The exchange to which the order should be submitted.
    ccxt_params : dict, optional
        The CCXT parameters dictionary to pass when creating an order. The
        default is {}.
    """

    def __init__(
        self,
        instrument: str = None,
        direction: int = None,
        order_type: str = "market",
        size: float = None,
        order_limit_price: float = None,
        order_stop_price: float = None,
        stop_loss: float = None,
        stop_type: str = None,
        take_profit: float = None,
        **kwargs,
    ) -> Order:

        # Required attributes
        self.instrument = instrument
        self.direction = direction
        self.order_type = order_type

        # Optional arguments
        self.size = size
        self.base_size = None
        self.target_value = None
        self.order_price = None
        self.order_time = None
        self.order_limit_price = order_limit_price
        self.order_stop_price = order_stop_price
        self.pip_value = None
        self.HCF = 1

        # Precision
        self.price_precision = 5
        self.size_precision = 5

        # Multi-exchange handling
        self.exchange = None

        # Stop loss arguments
        self.stop_type = stop_type
        self.stop_loss = stop_loss
        self.stop_distance = None

        # Take profit arguments
        self.take_profit = take_profit
        self.take_distance = None

        self.related_orders = None

        # Reduce only order
        self.parent_order = None  # Parent order ID
        self.reduce_only = False
        self.OCO = []  # One-cancels-other

        self.data_name = None  # When using custom DataStream

        # IB attributes
        self.currency = None
        self.secType = None
        self.contract_month = None
        self.localSymbol = None

        # Oanda attributes
        self.trigger_price = "DEFAULT"

        # dydx attributes
        self.post_only = False
        self.limit_fee = "0.015"

        # CCXT attributes
        self.ccxt_params = {}  # CCXT order parameters
        self.ccxt_order = {}  # CCXT native order structure

        self.reason = None

        self.strategy = None
        self.granularity = None
        self._sizing = None
        self._risk_pc = None

        # Meta-data
        self.id = None
        self.status = None  # options: pending -> open -> cancelled | filled

        # Unpack kwargs
        for item in kwargs:
            setattr(self, item, kwargs[item])

        # Enforce stop type
        if self.stop_loss is not None or self.stop_distance is not None:
            self.stop_type = self.stop_type if self.stop_type is not None else "limit"

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        if self.instrument is None:
            # Blank order
            return "Blank order"

        else:
            # Order constructed with instrument
            if self.size is not None:
                if self.direction is None:
                    return "Invalid order (direction not specified)"

                side = "buy" if self.direction > 0 else "sell"

                string = (
                    f"{round(self.size, self.size_precision)} "
                    + f"unit {self.instrument} {self.order_type} "
                    + f"{side} order"
                )

                # Append additional information
                if self.order_type == "limit":
                    if self.order_limit_price is None:
                        return "Invalid order (limit price not provided)"
                    string += f" @ {self.order_limit_price}"

                elif self.order_type == "stop-limit":
                    if self.order_limit_price is None:
                        return "Invalid order (limit price not provided)"
                    elif self.order_stop_price is None:
                        return "Invalid order (stop price not provided)"
                    string += f" @ {self.order_stop_price} / {self.order_limit_price}"

                elif self.order_type == "stop":
                    if self.order_stop_price is None:
                        return "Invalid order (stop price not provided)"
                    string += f" @ {self.order_stop_price}"

                return string

            else:
                # Size un-assigned
                return f"{self.instrument} {self.order_type} Order"

    def __call__(
        self,
        broker=None,
        order_price: float = None,
        order_time: datetime = datetime.now(),
        HCF: float = None,
        precision: dict = None,
    ) -> None:
        """Order object, called before submission to broker in
        autobot._qualify_orders.

        Parameters
        ----------
        broker : AutoTrader broker API instance, optional
            The broker-autotrader api instance. The default is None.
        order_price : float, optional
            The order price. The default is None.
        order_time : datetime, optional
            The time of the order. The default is datetime.now().
        HCF : float, optional
            The home conversion factor. The default is 1.
        precision : dict, optional
            A dictionary containing the precision for order size and price.
            The default is None.

        Returns
        -------
        None
            Calling an Order will ensure all information is present.
        """
        self.order_price = order_price if order_price else self.order_price
        self.order_time = order_time if order_time else self.order_time
        self.HCF = HCF if HCF is not None else self.HCF

        # Assign precisions
        if precision is not None:
            self.price_precision = precision["price"]
            self.size_precision = precision["size"]

        # Enforce size scalar
        self.size = abs(self.size) if self.size is not None else self.size

        if self.order_type not in ["close", "modify"]:
            self._set_working_price()
            self._calculate_exit_prices(broker)
            self._calculate_position_size(broker)

        self.status = "submitted"
        self.submitted = True

    def _set_working_price(self, order_price: float = None) -> None:
        """Sets the Orders' working price, for calculating exit targets.

        Parameters
        ----------
        order_price : float, optional
            The order price.

        Returns
        -------
        None
            The working price will be saved as a class attribute.
        """
        order_price = order_price if order_price is not None else self.order_price
        if self.order_type == "limit" or self.order_type == "stop-limit":
            self._working_price = round(self.order_limit_price, self.price_precision)
        else:
            if order_price is not None:
                self._working_price = round(order_price, self.price_precision)
            else:
                self._working_price = None

    def _calculate_exit_prices(self, broker=None, working_price: float = None) -> None:
        """Calculates the prices of the exit targets from the pip distance
        values.

        Parameters
        ----------
        broker : AutoTrader Broker Interface, optional
            The autotrade-broker instance. The default is None.
        working_price : float, optional
            The working price used to calculate amount risked. The default is
            None.

        Returns
        -------
        None
            The exit prices will be assigned to the order instance.
        """
        working_price = (
            round(working_price, self.price_precision)
            if working_price is not None
            else self._working_price
        )

        if broker is None:
            # No broker provided, create nominal utils instance
            utils = BrokerUtils()
        else:
            # Use broker-specific utilities
            utils = broker._utils

        pip_value = (
            self.pip_value
            if self.pip_value is not None
            else utils.get_pip_ratio(self.instrument)
        )

        # Calculate stop loss price
        if self.stop_loss is None and self.stop_distance is not None:
            # Stop loss provided as pip distance, convert to price
            stop_loss = (
                working_price - np.sign(self.direction) * self.stop_distance * pip_value
            )
            self.stop_loss = round(stop_loss, self.price_precision)

        if (
            self.stop_type == "trailing"
            and self.stop_distance is None
            and working_price is not None
        ):
            # Convert stop_loss price to stop_distance as well
            self.stop_distance = (
                np.sign(self.direction) * (working_price - self.stop_loss) / pip_value
            )

        # Calculate take profit price
        if self.take_profit is None and self.take_distance is not None:
            # Take profit pip distance specified, convert to price
            take_profit = (
                working_price + np.sign(self.direction) * self.take_distance * pip_value
            )
            self.take_profit = round(take_profit, self.price_precision)

    def _calculate_position_size(
        self,
        broker=None,
        working_price: float = None,
        HCF: float = 1,
        risk_pc: float = 0,
        sizing: str | float = "risk",
        amount_risked: float = None,
    ) -> None:
        """Calculates trade size for order.

        Parameters
        ----------
        broker : AutoTrader Broker Interface, optional
            The autotrade-broker instance. The default is None.
        working_price : float, optional
            The working price used to calculate amount risked. The default is None.
        HCF : float, optional
            The home conversion factor. The default is 1.
        risk_pc : float, optional
            The percentage of the account NAV to risk on the trade. The default is 0.
        sizing : str | float, optional
            The sizing option. The default is 'risk'.
        amount_risked : float, optional
            The dollar amount risked on the trade. The default is None.

        Returns
        -------
        None
            The trade size will be assigned to the order instance.
        """
        working_price = (
            working_price if working_price is not None else self._working_price
        )
        HCF = self.HCF if self.HCF is not None else HCF
        sizing = self._sizing if self._sizing is not None else sizing
        risk_pc = self._risk_pc if self._risk_pc is not None else risk_pc

        if self.size is None:
            # Size has not been set
            if self.target_value is not None:
                # Calculate size based on target trade value
                self.size = round(
                    self.target_value / working_price / HCF, self.size_precision
                )
            elif self.base_size is not None:
                # Size provided in base units, need to convert it
                self.size = round(self.base_size / HCF, self.size_precision)
            else:
                # Size not provided, need to calculate it
                if sizing == "risk":
                    # Calculate size from SL placement
                    try:
                        amount_risked = (
                            amount_risked
                            if amount_risked
                            else broker.get_NAV() * risk_pc / 100
                        )

                        size = broker._utils.get_size(
                            instrument=self.instrument,
                            amount_risked=amount_risked,
                            price=working_price,
                            HCF=HCF,
                            stop_price=self.stop_loss,
                            stop_distance=self.stop_distance,
                        )
                        self.size = round(size, self.size_precision)

                    except AttributeError:
                        # Broker is None type
                        raise Exception(
                            "Cannot calculate size without broker access. Please "
                            + "pass size argument, or add broker to Order object."
                        )

                else:
                    # Use position size provided via sizing key
                    self.size = sizing

        else:
            # Size has been set, enforce precision
            self.size = round(self.size, self.size_precision)

    @classmethod
    def _partial_fill(cls, order: Order, units_filled: float) -> Order:
        """Partially fills the order."""
        # Instantiate new order
        order_to_be_filled = cls()

        # Inherit attributes from base order
        for attribute, value in order.__dict__.items():
            setattr(order_to_be_filled, attribute, value)

        # Reset ID
        order_to_be_filled.id = None

        # Transfer units
        order_to_be_filled.size = units_filled
        order.size -= units_filled

        return order_to_be_filled

    def _check_precision(
        self,
    ):
        # TODO - implement
        raise NotImplementedError("This method has not been implemented yet.")

    def _validate(
        self,
    ):
        # TODO - add order validation method, ie. for IB, check all attributes are
        # assigned (eg. sectype, etc)
        raise NotImplementedError("This method has not been implemented yet.")

    def as_dict(self) -> dict:
        """Converts Order object to dictionary.

        Returns
        -------
        dict
            The order instance returned as a dict object.

        Notes
        -----
        This method enables legacy code operation, returning order/trade
        objects as a dictionary.
        """
        return self.__dict__

    @classmethod
    def _from_dict(cls, order_dict: dict) -> Order:
        return Order(**order_dict)


class MarketOrder(Order):
    """Market order type."""

    def __init__(
        self,
        instrument: str = None,
        direction: int = None,
        size: float = None,
        **kwargs,
    ):
        # Create base Order
        super().__init__(
            instrument=instrument,
            direction=direction,
            order_type="market",
            size=size,
            **kwargs,
        )


class LimitOrder(Order):
    """Limit order type."""

    def __init__(
        self,
        instrument: str = None,
        direction: int = None,
        size: float = None,
        order_limit_price: float = None,
        **kwargs,
    ):
        # Create base Order
        super().__init__(
            instrument=instrument,
            direction=direction,
            order_type="limit",
            size=size,
            order_limit_price=order_limit_price,
            **kwargs,
        )


class StopLimitOrder(Order):
    """Stop-limit order type."""

    def __init__(
        self,
        instrument: str = None,
        direction: int = None,
        size: float = None,
        order_limit_price: float = None,
        order_stop_price: float = None,
        **kwargs,
    ):
        # Create base Order
        super().__init__(
            instrument=instrument,
            direction=direction,
            order_type="stop-limit",
            size=size,
            order_limit_price=order_limit_price,
            order_stop_price=order_stop_price,
            **kwargs,
        )


class IsolatedPosition(Order):
    """AutoTrader IsolatedPosition. Use to connect SL and TP orders to individual
    trades.

    Attributes
    ----------
    unrealised_PL : float
        The floating PnL of the trade.
    margin_required : float
        The margin required to maintain the trade.
    time_filled : datetime
        The time at which the trade was filled.
    fill_price : float
        The price at which the trade was filled.
    last_price : float
        The last price observed for the instrument associated with the trade.
    last_time : datetime
        The last time observed for the instrument associated with the trade.
    exit_price : float
        The price at which the trade was closed.
    exit_time : datetime
        The time at which the trade was closed.
    fees : float
        The fees associated with the trade.
    parent_id : int
        The ID of the order which spawned the trade.
    id : int
        The trade ID.
    status : str
        The status of the trade (open or closed).
    split : bool
        If the trade has been split.

    Notes
    -----
    When a trade is created from an Order, the Order will be marked as filled.
    """

    def __init__(self, order: Order = None, **kwargs) -> IsolatedPosition:

        # Trade data
        self.unrealised_PL = 0
        self.margin_required = 0
        self.time_filled = None
        self.fill_price = None

        self.last_price = None
        self.last_time = None

        self.profit = 0
        self.balance = None
        self.exit_price = None
        self.exit_time = None
        self.fees = None

        # Meta data
        self.parent_id = None  # ID of order which spawned trade
        self.id = None
        self.status = None  # options: open -> closed
        self.split = False

        # Unpack kwargs
        for item in kwargs:
            setattr(self, item, kwargs[item])

        # Inherit order attributes
        if order:
            self._inheret_order(order)
            self.parent_id = order.id

    def __repr__(self):
        direction = "long" if self.direction > 0 else "short"
        return (
            f"{round(self.size,3)} unit {direction} {self.instrument} IsolatedPosition"
        )

    def __str__(self):
        return "AutoTrader IsolatedPosition"

    def _inheret_order(self, order: Order) -> None:
        for attribute, value in order.__dict__.items():
            setattr(self, attribute, value)

    @classmethod
    def _split(cls, trade: IsolatedPosition, split_units: float) -> IsolatedPosition:
        """Splits parent IsolatedPosition into new object for partial
        closures.

        split units are given to the new trade.
        """
        split_trade = cls()
        for attribute, value in trade.__dict__.items():
            setattr(split_trade, attribute, value)

        # Reset ID
        split_trade.parent_id = trade.parent_id
        split_trade.order_id = None

        # Transfer units
        split_trade.size = split_units
        trade.size -= split_units

        # Mark original trade as split
        trade.split = True

        return split_trade


class Trade:
    """AutoTrader Trade object. Represents an exchange of value."""

    def __init__(
        self,
        instrument: str,
        order_price: float,
        order_time: datetime,
        order_type: str,
        size: float,
        last_price: float,
        fill_time: datetime,
        fill_price: float,
        fill_direction: int,
        fee: float,
        **kwargs,
    ) -> Trade:
        """Trade constructor."""
        # Trade data
        self.fill_time = fill_time
        self.fill_price = fill_price
        self.direction = fill_direction
        self.fee = fee
        self.last_price = last_price
        self.order_price = order_price
        self.order_time = order_time
        self.order_type = order_type
        self.size = size
        self.instrument = instrument

        # Precision attributes
        self._price_precision = 5
        self._size_precision = 5

        # Meta-data
        self.id = None
        self.order_id = None

        for item in kwargs:
            setattr(self, item, kwargs[item])

    def __repr__(self):
        direction = "long" if self.direction > 0 else "short"
        return f"{round(self.size,3)} unit {direction} {self.instrument} trade @ {self.fill_price}"

    def __str__(self):
        return "AutoTrader Trade"


class Position:
    """AutoTrader Position object.

    Attributes
    ----------
    instrument : str
        The trade instrument of the position.
    pnl : float
        The pnl of the position.
    long_units : float
        The number of long units in the position.
    long_PL : float
        The PnL of the long units in the position.
    long_margin : float
        The margin required to maintain the long units in the position.
    short_units : float
        The number of short units in the position.
    short_PL : float
        The PnL of the short units in the position.
    short_margin : float
        The margin required to maintain the short units in the position.
    total_margin : float
        The total margin required to maintain the position.
    trade_IDs : list[int]
        The trade ID's associated with the position.
    net_position : float
        The total number of units in the position.
    net_exposure : float
        The net exposure (in $ value) of the position.
    PL : float
        The floating PnL (IB only).
    contracts : list
        The contracts associated with the position (IB only).
    portfolio_items : list
        The portfolio items associated with the position (IB only).
    """

    def __init__(self, **kwargs):
        self.instrument = None
        self.net_position = None
        self.pnl = 0
        self.long_units = None
        self.long_PL = None
        self.long_margin = None
        self.short_units = None
        self.short_PL = None
        self.short_margin = None
        self.total_margin = None
        self.trade_IDs = None
        self.net_exposure = None
        self.notional = 0
        self.price_precision = 5
        self.size_precision = 5
        self.avg_price = None
        self._prev_avg_price = None

        # TODO - include position open time, close time
        # Also need to track max units ? I think that can all be built
        # from fills
        self.HCF = 1

        self.entry_time = None
        self.last_price = None
        self.last_time = None
        self.exit_time = None
        self.exit_price = None

        self.direction = None
        # self.max_size = None # ?

        # IB Attributes
        self.PL = None
        self.contracts = None
        self.portfolio_items = None

        # dYdX Attributes
        self.entry_price = None

        # CCXT Attributes
        self.ccxt = None

        for item in kwargs:
            setattr(self, item, kwargs[item])

    def __repr__(self):
        return f"Position in {self.instrument}"

    def __str__(self):
        return "AutoTrader Position"

    def _update_with_fill(self, trade: Trade):
        """Updates the position with the order provided."""
        net_position_before_fill = self.net_position

        # Update net position
        self.net_position += trade.size * trade.direction

        # Update average entry price
        self._prev_avg_price = self.avg_price
        if self.net_position * net_position_before_fill >= 0:
            # Position has not flipped
            if abs(self.net_position) > abs(net_position_before_fill):
                # Position has increased
                self.avg_price = round(
                    (
                        self.avg_price * abs(self.net_position)
                        + trade.fill_price * trade.size
                    )
                    / (abs(self.net_position) + trade.size),
                    self.price_precision,
                )
            # If position has reduced, average entry price will not change
        else:
            # Position has flipped
            self.avg_price = round(trade.fill_price, self.price_precision)

        # Update last price and last time
        self.last_price = trade.last_price
        self.last_time = trade.fill_time

        # TODO - update value of position
        # self.net_exposure
        self.notional = self.last_price * abs(self.net_position)

    @classmethod
    def _from_fill(cls, trade: Trade):
        """Returns a Position from a fill."""
        position = cls(
            instrument=trade.instrument,
            net_position=abs(trade.size) * trade.direction,
            last_price=trade.last_price,
            last_time=trade.fill_time,
            entry_time=trade.fill_time,
            entry_price=trade.fill_price,
            avg_price=trade.fill_price,
            notional=trade.fill_price * abs(trade.size),
            price_precision=trade._price_precision,
            size_precision=trade._size_precision,
            direction=trade.direction,
        )
        # TODO - update attributes created
        # - exposure, value, etc.
        return position

    def as_dict(self) -> dict:
        """Converts Position object to dictionary.

        Returns
        -------
        dict
            The Position instance returned as a dict object.

        Notes
        -----
        This method enables legacy code operation, returning order/trade
        objects as a dictionary.
        """
        return self.__dict__



import setuptools
import codecs
import os.path


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


# Define extra dependencies
dydx_dep = ["dydx-v3-python >= 1.9.0"]
ccxt_dep = ["ccxt >= 2.0.53"]
oanda_dep = [
    "v20 >= 3.0.25.0",
]
ib_dep = [
    "ib_insync >= 0.9.70",
]
yfinance_dep = [
    "yfinance >= 0.1.67",
]
dev_dep = [
    "pytest >= 7.1.1",
    "black >= 22.10.0",
    "sphinx-copybutton >= 0.5.0",
    "sphinx-inline-tabs >= 2022.1.2b11",
    "myst-parser >= 0.18.1",
    "furo >= 2022.9.29",
    "sphinx-autobuild >= 2021.3.14",
    "commitizen >= 2.35.0",
]
all_dep = ccxt_dep + oanda_dep + ib_dep + yfinance_dep + dev_dep

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="autotrader",
    version=get_version("autotrader/__init__.py"),
    author="Kieran Mackle",
    author_email="kemackle98@gmail.com",
    license="gpl-3.0",
    description="A Python-based platform for developing, optimising "
    + "and deploying automated trading systems.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://kieran-mackle.github.io/AutoTrader/",
    project_urls={
        "Bug Tracker": "https://github.com/kieran-mackle/AutoTrader/issues",
        "Source Code": "https://github.com/kieran-mackle/AutoTrader",
        "Documentation": "https://autotrader.readthedocs.io/en/latest/",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    keywords=["algotrading", "finance", "crypto", "forex", "python"],
    package_dir={"": "."},
    packages=setuptools.find_packages(where="."),
    python_requires=">=3.7",
    install_requires=[
        "numpy >= 1.20.3",
        "pandas >= 1.3.4",
        "art >= 5.7",
        "PyYAML",
        "bokeh >= 2.3.1",
        "scipy >= 1.7.1",
        "finta >= 1.3",
        "tqdm>=4.64.0",
        "importlib-resources",
        "click >= 8.1.3",
        "requests >= 2.28.1",
        "python-telegram-bot >= 13.14",
    ],
    extras_require={
        "dydx": dydx_dep,
        "ccxt": ccxt_dep,
        "oanda": oanda_dep,
        "ib": ib_dep,
        "yfinance": yfinance_dep,
        "dev": dev_dep,
        "all": all_dep,
    },
    setup_requires=[
        "setuptools_git",
        "setuptools_scm",
    ],
    package_data={"": ["data/*.js", "data/keys.yaml"]},
    entry_points={
        "console_scripts": [
            "autotrader = autotrader.bin.cli:cli",
        ],
    },
)



from abc import ABC, abstractmethod
from autotrader.brokers.trading import Order


class Notifier(ABC):
    @abstractmethod
    def __init__(self, *args, **kwargs) -> None:
        pass

    @abstractmethod
    def send_order(self, order: Order, *args, **kwargs) -> None:
        pass

    @abstractmethod
    def send_message(self, message: str, *args, **kwargs) -> None:
        pass



import sys
import yaml
import pickle
import numpy as np
import pandas as pd
from datetime import datetime, timedelta


def read_yaml(file_path: str) -> dict:
    """Function to read and extract contents from .yaml file.

    Parameters
    ----------
    file_path : str
        The absolute filepath to the yaml file.

    Returns
    -------
    dict
        The loaded yaml file in dictionary form.
    """
    with open(file_path, "r") as f:
        return yaml.safe_load(f)


def write_yaml(data: dict, filepath: str) -> None:
    """Writes a dictionary to a yaml file.

    Parameters
    ----------
    data : dict
        The dictionary to write to yaml.
    filepath : str
        The filepath to save the yaml file.

    Returns
    -------
    None
        The data will be written to the filepath provided.
    """
    with open(filepath, "w") as outfile:
        yaml.dump(data, outfile, default_flow_style=False)


def get_broker_config(
    broker: str, global_config: dict = None, environment: str = "paper"
) -> dict:
    """Returns a broker configuration dictionary.

    Parameters
    ----------
    broker : str
        The name(s) of the broker/exchange. Specify multiple exchanges using
        comma separation.
    global_config : dict
        The global configuration dictionary.
    environment : str, optional
        The trading evironment ('demo' or 'real').

    """
    all_config = {}
    inputted_brokers = broker.lower().replace(" ", "").split(",")

    # Check global_config
    if global_config is None:
        try:
            global_config = read_yaml("config/keys.yaml")
        except:
            pass

    for broker in inputted_brokers:
        # Check for CCXT
        if broker.split(":")[0].lower() == "ccxt":
            broker_key = broker
            broker, exchange = broker.lower().split(":")
        elif broker.split(":")[0].lower() == "virtual":
            broker_key = ":".join(broker.split(":")[1:])
            broker = "virtual"
        else:
            broker_key = broker

        supported_brokers = ["oanda", "ib", "ccxt", "dydx", "virtual"]
        if broker.lower() not in supported_brokers:
            raise Exception(f"Unsupported broker: '{broker}'")

        if broker != "ccxt" and environment.lower() not in ["live", "paper"]:
            raise Exception("Trading environment must either be 'live' or 'paper'.")

        # Live trading
        if broker.lower() == "oanda":
            api_key = "LIVE" if environment.lower() == "live" else "PRACTICE"
            oanda_conf = global_config["OANDA"]

            # Initialise config dict
            config = {"PORT": oanda_conf["PORT"]}

            # Unpack
            if f"{api_key}_API" in oanda_conf:
                config["API"] = oanda_conf[f"{api_key}_API"]
            else:
                raise Exception(
                    f"Please define {api_key}_API in your "
                    + f"account configuration for {environment} trading."
                )

            if f"{api_key}_ACCESS_TOKEN" in oanda_conf:
                config["ACCESS_TOKEN"] = oanda_conf[f"{api_key}_ACCESS_TOKEN"]
            else:
                raise Exception(
                    f"Please define {api_key}_ACCESS_TOKEN in "
                    + f"your account configuration for {environment} trading."
                )

            config["ACCOUNT_ID"] = (
                oanda_conf["DEFAULT_ACCOUNT_ID"]
                if "custom_account_id" not in global_config
                else global_config["custom_account_id"]
            )

        elif broker.lower() == "ib":
            config = {
                "host": global_config["host"]
                if "host" in global_config
                else "127.0.0.1",
                "port": global_config["port"] if "port" in global_config else 7497,
                "clientID": global_config["clientID"]
                if "clientID" in global_config
                else 1,
                "account": global_config["account"]
                if "account" in global_config
                else "",
                "read_only": global_config["read_only"]
                if "read_only" in global_config
                else False,
            }

        elif broker.lower() == "dydx":
            try:
                eth_address = global_config["DYDX"]["ETH_ADDRESS"]
                eth_private_key = global_config["DYDX"]["ETH_PRIV_KEY"]
                config = {
                    "data_source": "dYdX",
                    "ETH_ADDRESS": eth_address,
                    "ETH_PRIV_KEY": eth_private_key,
                }
            except KeyError:
                raise Exception(
                    "Using dYdX for trading requires authentication via "
                    + "the global configuration. Please make sure you provide the "
                    + "following keys:\n ETH_ADDRESS: your ETH address "
                    + "\n ETH_PRIV_KEY: your ETH private key."
                    + "These must all be provided under the 'dYdX' key."
                )

        elif broker.lower() == "ccxt":
            try:
                config_data = global_config[broker_key.upper()]

                # Select config based on environment
                if environment.lower() in config_data:
                    config_data = config_data[environment.lower()]
                elif "mainnet" in config_data and environment.lower() == "live":
                    config_data = config_data["mainnet"]
                elif "testnet" in config_data and environment.lower() == "paper":
                    config_data = config_data["testnet"]

                api_key = config_data["api_key"] if "api_key" in config_data else None
                secret = config_data["secret"] if "secret" in config_data else None
                currency = (
                    config_data["base_currency"]
                    if "base_currency" in config_data
                    else "USDT"
                )
                sandbox_mode = False if environment.lower() == "live" else True
                config = {
                    "data_source": "ccxt",
                    "exchange": exchange,
                    "api_key": api_key,
                    "secret": secret,
                    "sandbox_mode": sandbox_mode,
                    "base_currency": currency,
                }
                other_args = {"options": {}, "password": None}
                for key, default_val in other_args.items():
                    if key in config_data:
                        config[key] = config_data[key]
                    else:
                        config[key] = default_val

            except KeyError:
                raise Exception(
                    "Using CCXT for trading requires authentication via "
                    + "the global configuration. Please make sure you provide the "
                    + "details in the following format:\n"
                    + "CCXT:EXCHANGE_NAME:\n"
                    + '  api_key: "xxxx" (the exchange-specific api key)\n'
                    + '  secret: "xxxx" (the exchange-specific api secret)\n'
                    + "  base_currency: USDT (your account's base currency)\n"
                )

        elif broker.lower() == "virtual":
            config = {}

        else:
            raise Exception(f"No configuration available for {broker}.")

        # Append to full config
        all_config[broker_key] = config

    # Check length
    if len(all_config) == 1:
        all_config = config

    return all_config


def get_data_config(feed: str, global_config: dict = None, **kwargs) -> dict:
    """Returns a configuration dictionary for AutoData.
    Parameters
    ----------
    global_config : dict
        The global configuration dictionary.
    feed : str
        The name of the data feed.
    """
    if feed is None:
        print("Please specify a data feed.")
        sys.exit(0)

    # Check for CCXT
    if feed.split(":")[0].lower() == "ccxt":
        feed, exchange = feed.lower().split(":")

    # Check feed
    supported_feeds = ["oanda", "ib", "ccxt", "dydx", "yahoo", "local", "none"]
    if feed.lower() not in supported_feeds:
        raise Exception(f"Unsupported data feed: '{feed}'")

    # Check global_config
    if global_config is None:
        try:
            global_config = read_yaml("config/keys.yaml")
        except:
            pass

    # Check for required authentication
    auth_feeds = ["oanda", "ib"]
    if feed.lower() in auth_feeds and global_config is None:
        raise Exception(
            f"Data feed '{feed}' requires authentication. "
            + "Please provide authentication details in the global config."
        )

    # Construct configuration dict
    config = {"data_source": feed.lower()}

    if feed.lower() == "oanda":
        environment = kwargs["environment"] if "environment" in kwargs else "paper"
        api_key = "LIVE" if environment.lower() == "live" else "PRACTICE"
        oanda_conf = global_config["OANDA"]

        # Unpack
        if f"{api_key}_API" in oanda_conf:
            config["API"] = oanda_conf[f"{api_key}_API"]
        else:
            raise Exception(
                f"Please define {api_key}_API in your "
                + f"account configuration for {environment} trading."
            )

        if f"{api_key}_ACCESS_TOKEN" in oanda_conf:
            config["ACCESS_TOKEN"] = oanda_conf[f"{api_key}_ACCESS_TOKEN"]
        else:
            raise Exception(
                f"Please define {api_key}_ACCESS_TOKEN in "
                + f"your account configuration for {environment} trading."
            )

        config["PORT"] = oanda_conf["PORT"]
        config["ACCOUNT_ID"] = (
            oanda_conf["DEFAULT_ACCOUNT_ID"]
            if "custom_account_id" not in global_config
            else global_config["custom_account_id"]
        )

    elif feed.lower() == "ib":
        config["host"] = (
            global_config["host"] if "host" in global_config else "127.0.0.1"
        )
        config["port"] = global_config["port"] if "port" in global_config else 7497
        config["clientID"] = (
            global_config["clientID"] if "clientID" in global_config else 1
        )
        config["account"] = (
            global_config["account"] if "account" in global_config else ""
        )
        config["read_only"] = (
            global_config["read_only"] if "read_only" in global_config else False
        )

    elif feed.lower() == "ccxt":
        # Try add authentication with global config
        if global_config is not None:
            # Global config is available
            try:
                # Try get api keys
                environment = (
                    kwargs["environment"] if "environment" in kwargs else "paper"
                )
                config = get_broker_config(
                    broker=f"{feed.lower()}:{exchange}",
                    global_config=global_config,
                    environment=environment,
                )
            except:
                # Didn't work, just set exchange
                config["exchange"] = exchange
        else:
            # No global config available, just set exchange
            config["exchange"] = exchange

    return config


def get_streaks(trade_summary):
    """Calculates longest winning and losing streaks from trade summary."""
    profit_list = trade_summary[trade_summary["status"] == "closed"].profit.values
    longest_winning_streak = 1
    longest_losing_streak = 1
    streak = 1

    for i in range(1, len(profit_list)):
        if np.sign(profit_list[i]) == np.sign(profit_list[i - 1]):
            streak += 1

            if np.sign(profit_list[i]) > 0:
                # update winning streak
                longest_winning_streak = max(longest_winning_streak, streak)
            else:
                # Update losing
                longest_losing_streak = max(longest_losing_streak, streak)

        else:
            streak = 1

    return longest_winning_streak, longest_losing_streak


def unpickle_broker(picklefile: str = ".virtual_broker"):
    """Unpickles a virtual broker instance for post-processing."""
    with open(picklefile, "rb") as file:
        instance = pickle.load(file)
    return instance


class TradeAnalysis:
    """AutoTrader trade analysis class.

    Attributes
    ----------
    instruments_traded : list
        The instruments traded during the trading period.
    account_history : pd.DataFrame
        A timeseries history of the account during the trading period.
    holding_history : pd.DataFrame
        A timeseries summary of holdings during the trading period, by portfolio
        allocation fraction.
    order_history : pd.DataFrame
        A timeseries history of orders placed during the trading period.
    cancelled_orders : pd.DataFrame
        Orders which were cancelled during the trading period.
    trade_history : pd.DataFrame
        A history of all trades (fills) made during the trading period.

    """

    def __init__(
        self,
        broker,
        broker_histories: dict,
        instrument: str = None,
        price_history: pd.DataFrame = None,
    ):
        # Meta data
        self.brokers_used = None
        self.broker_results = None
        self.instruments_traded = None

        # Histories
        self.account_history = None
        self.position_history = None
        self.position_summary = None
        self.order_history = None
        self.cancelled_orders = None
        self.trade_history = None
        self.price_history = price_history

        # Perform analysis
        self.analyse_account(broker, broker_histories, instrument)

    def __str__(self):
        return "AutoTrader Trading Results"

    def __repr__(self):
        return "AutoTrader Trading Results"

    def analyse_account(
        self,
        broker,
        broker_histories: dict,
        instrument: str = None,
    ) -> None:
        """Analyses trade account and creates summary of key details."""
        if not isinstance(broker, dict):
            # Single broker - create dummy dict
            broker_instances = {list(broker_histories.keys())[0]: broker}
        else:
            # Multiple brokers passed in as dict
            broker_instances = broker

        # Process results from each broker instance
        broker_results = {}
        for broker_name, broker in broker_instances.items():
            # Construct trade and order summaries
            all_orders = {}
            for status in ["pending", "open", "filled", "cancelled"]:
                orders = broker.get_orders(order_status=status)
                all_orders.update(orders)

            orders = TradeAnalysis.create_trade_summary(
                orders=all_orders, instrument=instrument, broker_name=broker_name
            )
            trade_history = TradeAnalysis.create_fill_summary(
                fills=broker._fills, broker_name=broker_name
            )

            account_history = pd.DataFrame(
                data=broker_histories[broker_name],
            )
            account_history.set_index("time", inplace=True)

            position_history = TradeAnalysis.create_position_history(
                trade_history=trade_history,
                account_history=account_history,
            )

            position_summary = TradeAnalysis._create_position_summary(
                broker._positions, broker._closed_positions
            )

            # Calculate drawdown
            account_history["drawdown"] = (
                account_history.NAV / account_history.NAV.cummax() - 1
            )

            # Save results for this broker instance
            broker_results[broker_name] = {
                "instruments_traded": list(orders.instrument.unique()),
                "account_history": account_history,
                "position_history": position_history,
                "position_summary": position_summary,
                "order_history": orders,
                "cancelled_orders": orders[orders.status == "cancelled"],
                "trade_history": trade_history,
            }

        # Save all results
        self.broker_results = broker_results

        # Aggregate across broker instances
        self._aggregate_across_brokers(broker_results)

    @staticmethod
    def create_position_history(
        trade_history: pd.DataFrame,
        account_history: pd.DataFrame,
    ) -> pd.DataFrame:
        """Creates a history of positions held, recording number of units held
        at each timestamp.
        """
        # Use fills to reconstruct position history, in terms of units held
        instruments_traded = list(trade_history["instrument"].unique())

        position_histories_dict = {}
        for instrument in instruments_traded:
            instrument_trade_hist = trade_history[
                trade_history["instrument"] == instrument
            ]
            directional_trades = (
                instrument_trade_hist["direction"] * instrument_trade_hist["size"]
            )
            net_position_hist = round(directional_trades.cumsum(), 8)

            # Filter out duplicates
            net_position_hist = net_position_hist[
                ~net_position_hist.index.duplicated(keep="last")
            ]

            # Reindex to account history index
            # TODO - match index timezone info
            net_position_hist = net_position_hist[~net_position_hist.index.isna()]
            try:
                net_position_hist = net_position_hist.reindex(
                    index=account_history.index, method="ffill"
                ).fillna(0)
            except:
                print("Could not reindex position history on account history index.")

            # Save result
            position_histories_dict[instrument] = net_position_hist

        position_histories = pd.concat(position_histories_dict, axis=1)
        return position_histories

    @staticmethod
    def _create_position_summary(open_positions, closed_positions):
        """Creates a summary of positions held."""
        # # Analyse closed positions (currently unused)
        # open_positions_summary = {}
        # for instrument, position in open_positions.items():
        #     directions = [p.direction for p in positions]

        #     # Save analysis
        #     open_positions_summary[instrument] = {
        #         "directions": directions,
        #         "no_long": directions.count(1),
        #         "no_short": directions.count(-1),
        #     }

        # Analyse closed positions
        closed_positions_results = {}
        closed_positions_summary = {}
        for instrument, positions in closed_positions.items():
            directions = [p.direction for p in positions]
            durations = [p.exit_time - p.entry_time for p in positions]

            # Save analysis
            closed_positions_results[instrument] = {
                "directions": directions,
                "durations": durations,
            }
            closed_positions_summary[instrument] = {
                "no_long": directions.count(1),
                "no_short": directions.count(-1),
            }
            if len(durations) > 0:
                closed_positions_summary[instrument]["avg_duration"] = np.mean(
                    durations
                )
                if directions.count(1) > 0:
                    closed_positions_summary[instrument]["avg_long_duration"] = np.mean(
                        np.array(durations)[np.array(directions) == 1]
                    )
                else:
                    closed_positions_summary[instrument]["avg_long_duration"] = None
                if directions.count(-1) > 0:
                    closed_positions_summary[instrument][
                        "avg_short_duration"
                    ] = np.mean(np.array(durations)[np.array(directions) == -1])
                else:
                    closed_positions_summary[instrument]["avg_short_duration"] = None

        # Create summary dataframe
        summary = pd.DataFrame(
            data=closed_positions_summary,
        )

        return summary

    def _aggregate_across_brokers(self, broker_results):
        """Aggregates trading history across all broker instances."""
        brokers_used = []
        instruments_traded = []
        cancelled_orders = pd.DataFrame()
        position_history = pd.DataFrame()
        position_summary = pd.DataFrame()
        order_history = pd.DataFrame()
        trade_history = pd.DataFrame()
        account_history = None
        for broker, results in broker_results.items():
            orders = results["order_history"]
            brokers_used.append(broker)

            # Append unique instruments traded
            unique_instruments = orders.instrument.unique()
            [
                instruments_traded.append(instrument)
                if instrument not in instruments_traded
                else None
                for instrument in unique_instruments
            ]

            # Aggregate account history
            if account_history is None:
                # Initialise
                account_history = results["account_history"]
            else:
                # Reindex each dataset
                original_index = results["account_history"].reindex(
                    index=account_history.index, method="ffill"
                )
                new_index = account_history.reindex(
                    index=results["account_history"].index, method="ffill"
                )
                if len(original_index) >= len(new_index):
                    # Use original index
                    account_history += original_index
                else:
                    # Use new index
                    account_history = new_index + results["account_history"]

            # Concatenate trades, orders and trade_history
            cancelled_orders = pd.concat(
                [cancelled_orders, results["cancelled_orders"]]
            )
            order_history = pd.concat([order_history, results["order_history"]])
            trade_history = pd.concat([trade_history, results["trade_history"]])
            position_history = pd.concat(
                [position_history, results["position_history"]]
            )
            # TODO - implement position summary for multiple exchanges
            position_summary = results["position_summary"]

        # Assign attributes
        self.brokers_used = brokers_used
        self.instruments_traded = instruments_traded
        self.account_history = account_history
        self.position_history = position_history
        self.position_summary = position_summary
        self.order_history = order_history
        self.cancelled_orders = cancelled_orders
        self.trade_history = trade_history

    @staticmethod
    def create_fill_summary(fills: list, broker_name: str = None):
        """Creates a dataframe of fill history."""
        # Initialise lists
        fill_dict = {
            "order_time": [],
            "order_price": [],
            "order_type": [],
            "fill_time": [],
            "fill_price": [],
            "direction": [],
            "size": [],
            "fee": [],
            "instrument": [],
            "id": [],
            "order_id": [],
        }
        for fill in fills:
            fill_dict["order_time"].append(fill.order_time)
            fill_dict["order_price"].append(fill.order_price)
            fill_dict["fill_time"].append(fill.fill_time)
            fill_dict["fill_price"].append(fill.fill_price)
            fill_dict["direction"].append(fill.direction)
            fill_dict["size"].append(fill.size)
            fill_dict["fee"].append(fill.fee)
            fill_dict["instrument"].append(fill.instrument)
            fill_dict["id"].append(fill.id)
            fill_dict["order_id"].append(fill.order_id)
            fill_dict["order_type"].append(fill.order_type)

        fill_df = pd.DataFrame(data=fill_dict, index=fill_dict["fill_time"])
        fill_df["broker"] = broker_name

        return fill_df

    @staticmethod
    def create_trade_summary(
        trades: dict = None,
        orders: dict = None,
        instrument: str = None,
        broker_name: str = None,
    ) -> pd.DataFrame:
        """Creates a summary dataframe for trades and orders."""
        instrument = None if isinstance(instrument, list) else instrument

        if trades is not None:
            iter_dict = trades
        else:
            iter_dict = orders

        iter_dict = {} if iter_dict is None else iter_dict

        product = []
        status = []
        ids = []
        times_list = []
        order_price = []
        size = []
        direction = []
        stop_price = []
        take_price = []
        reason = []

        if trades is not None:
            entry_time = []
            fill_price = []
            profit = []
            portfolio_balance = []
            exit_times = []
            exit_prices = []
            trade_duration = []
            fees = []

        for ID, item in iter_dict.items():
            product.append(item.instrument)
            status.append(item.status)
            ids.append(item.id)
            size.append(item.size)
            direction.append(item.direction)
            times_list.append(item.order_time)
            order_price.append(item.order_price)
            stop_price.append(item.stop_loss)
            take_price.append(item.take_profit)
            reason.append(item.reason)

        if trades is not None:
            for trade_id, trade in iter_dict.items():
                entry_time.append(trade.time_filled)
                fill_price.append(trade.fill_price)
                profit.append(trade.profit)
                portfolio_balance.append(trade.balance)
                exit_times.append(trade.exit_time)
                exit_prices.append(trade.exit_price)
                fees.append(trade.fees)
                if trade.status == "closed":
                    if type(trade.exit_time) == str:
                        exit_dt = datetime.strptime(
                            trade.exit_time, "%Y-%m-%d %H:%M:%S%z"
                        )
                        entry_dt = datetime.strptime(
                            trade.time_filled, "%Y-%m-%d %H:%M:%S%z"
                        )
                        trade_duration.append(
                            exit_dt.timestamp() - entry_dt.timestamp()
                        )
                    elif isinstance(trade.exit_time, pd.Timestamp):
                        trade_duration.append(
                            (trade.exit_time - trade.time_filled).total_seconds()
                        )
                    elif trade.exit_time is None:
                        # Weird edge case
                        trade_duration.append(None)
                    else:
                        trade_duration.append(
                            trade.exit_time.timestamp() - trade.time_filled.timestamp()
                        )
                else:
                    trade_duration.append(None)

            dataframe = pd.DataFrame(
                {
                    "instrument": product,
                    "status": status,
                    "ID": ids,
                    "order_price": order_price,
                    "order_time": times_list,
                    "fill_time": entry_time,
                    "fill_price": fill_price,
                    "size": size,
                    "direction": direction,
                    "stop_loss": stop_price,
                    "take_profit": take_price,
                    "profit": profit,
                    "balance": portfolio_balance,
                    "exit_time": exit_times,
                    "exit_price": exit_prices,
                    "trade_duration": trade_duration,
                    "fees": fees,
                },
                index=pd.to_datetime(entry_time),
            )

            # Fill missing values for balance
            dataframe.balance.fillna(method="ffill", inplace=True)

        else:
            # Order summary
            dataframe = pd.DataFrame(
                {
                    "instrument": product,
                    "status": status,
                    "order_id": ids,
                    "order_price": order_price,
                    "order_time": times_list,
                    "size": size,
                    "direction": direction,
                    "stop_loss": stop_price,
                    "take_profit": take_price,
                    "reason": reason,
                },
                index=pd.to_datetime(times_list),
            )

        dataframe = dataframe.sort_index()

        # Add broker name column
        dataframe["broker"] = broker_name

        # Filter by instrument
        if instrument is not None:
            dataframe = dataframe[dataframe["instrument"] == instrument]

        return dataframe

    def summary(self) -> dict:
        """Constructs a trading summary for printing."""
        # Initialise trade results dict
        trade_results = {}

        # Analyse account history
        if len(self.account_history) > 0:
            # The account was open for some nonzero time period
            starting_balance = self.account_history.equity[0]
            ending_balance = self.account_history.equity[-1]
            ending_NAV = self.account_history.NAV[-1]
            abs_return = ending_balance - starting_balance
            pc_return = 100 * abs_return / starting_balance
            floating_pnl = ending_NAV - ending_balance
            max_drawdown = min(self.account_history.drawdown)

            # Save results to dict
            trade_results["start"] = self.account_history.index[0]
            trade_results["end"] = self.account_history.index[-1]
            trade_results["starting_balance"] = starting_balance
            trade_results["ending_balance"] = ending_balance
            trade_results["ending_NAV"] = ending_NAV
            trade_results["abs_return"] = abs_return
            trade_results["pc_return"] = pc_return
            trade_results["floating_pnl"] = floating_pnl
            trade_results["max_drawdown"] = max_drawdown

        # All trades
        no_trades = len(self.trade_history)
        trade_results["no_trades"] = no_trades
        trade_results["no_long_trades"] = len(
            self.trade_history[self.trade_history["direction"] > 0]
        )
        trade_results["no_short_trades"] = len(
            self.trade_history[self.trade_history["direction"] < 0]
        )

        if no_trades > 0:
            # Initialise all_trades dict
            trade_results["all_trades"] = {}

            # Calculate positions still open
            # TODO - debug below with multi asset backtest
            try:
                trade_results["no_open"] = sum(self.position_history.iloc[-1] > 0)
            except:
                trade_results["no_open"] = 0

            # Analyse winning positions
            # wins = self.isolated_position_history[
            #     self.isolated_position_history.profit > 0
            # ]
            # avg_win = np.mean(wins.profit)
            # max_win = np.max(wins.profit)

            # Analyse losing positions
            # loss = self.isolated_position_history[
            #     self.isolated_position_history.profit < 0
            # ]
            # avg_loss = abs(np.mean(loss.profit))
            # max_loss = abs(np.min(loss.profit))

            # Performance
            # win_rate = 100 * len(wins) / no_trades
            # longest_win_streak, longest_lose_streak = get_streaks(
            #     self.isolated_position_history
            # )
            # try:
            #     avg_trade_duration = np.nanmean(
            #         self.isolated_position_history.trade_duration.values
            #     )
            #     trade_results["all_trades"]["avg_trade_duration"] = str(
            #         timedelta(seconds=int(avg_trade_duration))
            #     )
            # except TypeError:
            #     # Position has not been closed yet
            #     trade_results["all_trades"]["avg_trade_duration"] = None

            # Trade durations
            # duration_list = [
            #     i
            #     for i in self.isolated_position_history.trade_duration.values
            #     if i is not None
            # ]
            # if len(duration_list) > 0:
            #     min_trade_duration = np.nanmin(duration_list)
            #     max_trade_duration = np.nanmax(duration_list)
            # else:
            #     min_trade_duration = None
            #     max_trade_duration = None

            # Fees
            total_fees = self.trade_history.fee.sum()

            # Volume traded
            total_volume = (
                self.trade_history["size"] * self.trade_history["fill_price"]
            ).values.sum()

            # trade_results["all_trades"]["avg_win"] = avg_win
            # trade_results["all_trades"]["max_win"] = max_win
            # trade_results["all_trades"]["avg_loss"] = avg_loss
            # trade_results["all_trades"]["max_loss"] = max_loss
            # trade_results["all_trades"]["win_rate"] = win_rate
            # trade_results["all_trades"]["win_streak"] = longest_win_streak
            # trade_results["all_trades"]["lose_streak"] = longest_lose_streak
            trade_results["all_trades"]["avg_win"] = 0
            trade_results["all_trades"]["max_win"] = 0
            trade_results["all_trades"]["avg_loss"] = 0
            trade_results["all_trades"]["max_loss"] = 0
            trade_results["all_trades"]["win_rate"] = 0
            trade_results["all_trades"]["win_streak"] = 0
            trade_results["all_trades"]["lose_streak"] = 0
            trade_results["all_trades"]["total_volume"] = total_volume
            max_trade_duration = None
            min_trade_duration = None

            if max_trade_duration is not None:
                trade_results["all_trades"]["longest_trade"] = str(
                    timedelta(seconds=int(max_trade_duration))
                )
            else:
                trade_results["all_trades"]["longest_trade"] = str(None)

            if min_trade_duration is not None:
                trade_results["all_trades"]["shortest_trade"] = str(
                    timedelta(seconds=int(min_trade_duration))
                )
            else:
                trade_results["all_trades"]["shortest_trade"] = str(None)

            trade_results["all_trades"]["total_fees"] = total_fees

        # Cancelled orders
        trade_results["no_cancelled"] = len(self.cancelled_orders)

        # Long positions
        # long_positions = self.isolated_position_history[
        #     self.isolated_position_history["direction"] > 0
        # ]
        # no_long = len(long_positions)
        # trade_results["long_positions"] = {}
        # trade_results["long_positions"]["total"] = no_long
        # if no_long > 0:
        #     long_wins = long_positions[long_positions.profit > 0]
        #     avg_long_win = np.mean(long_wins.profit)
        #     max_long_win = np.max(long_wins.profit)
        #     long_loss = long_positions[long_positions.profit < 0]
        #     avg_long_loss = abs(np.mean(long_loss.profit))
        #     max_long_loss = abs(np.min(long_loss.profit))
        #     long_wr = 100 * len(long_positions[long_positions.profit > 0]) / no_long

        #     trade_results["long_positions"]["avg_long_win"] = avg_long_win
        #     trade_results["long_positions"]["max_long_win"] = max_long_win
        #     trade_results["long_positions"]["avg_long_loss"] = avg_long_loss
        #     trade_results["long_positions"]["max_long_loss"] = max_long_loss
        #     trade_results["long_positions"]["long_wr"] = long_wr

        # # Short positions
        # short_positions = self.isolated_position_history[
        #     self.isolated_position_history["direction"] < 0
        # ]
        # no_short = len(short_positions)
        # trade_results["short_positions"] = {}
        # trade_results["short_positions"]["total"] = no_short
        # if no_short > 0:
        #     short_wins = short_positions[short_positions.profit > 0]
        #     avg_short_win = np.mean(short_wins.profit)
        #     max_short_win = np.max(short_wins.profit)
        #     short_loss = short_positions[short_positions.profit < 0]
        #     avg_short_loss = abs(np.mean(short_loss.profit))
        #     max_short_loss = abs(np.min(short_loss.profit))
        #     short_wr = 100 * len(short_positions[short_positions.profit > 0]) / no_short

        #     trade_results["short_positions"]["avg_short_win"] = avg_short_win
        #     trade_results["short_positions"]["max_short_win"] = max_short_win
        #     trade_results["short_positions"]["avg_short_loss"] = avg_short_loss
        #     trade_results["short_positions"]["max_short_loss"] = max_short_loss
        #     trade_results["short_positions"]["short_wr"] = short_wr

        return trade_results


class DataStream:
    """Data stream class.

    This class is intended to provide a means of custom data pipelines.

    Methods
    -------
    refresh
        Returns up-to-date data, multi_data, quote_data and auxdata.
    get_trading_bars
        Returns a dictionary of the current bars for the products being
        traded, used to act on trading signals.

    Attributes
    ----------
    instrument : str
        The instrument being traded.
    feed : str
        The data feed.
    data_filepaths : str|dict
        The filepaths to locally stored data.
    quote_data_file : str
        The filepaths to locally stored quote data.
    auxdata_files : dict
        The auxiliary data files.
    strategy_params : dict
        The strategy parameters.
    get_data : AutoData
        The AutoData instance.
    data_start : datetime
        The backtest start date.
    data_end : datetime
        The backtest end date.
    portfolio : bool|list
        The instruments being traded in a portfolio, if any.
    data_path_mapper : callable
        A callable to map an instrument to an absolute filepath of
        data for that instrument.

    Notes
    -----
    A 'dynamic' dataset is one where the specific products being traded
    change over time. For example, trading contracts on an underlying product.
    In this case, dynamic_data should be set to True in AutoTrader.add_data
    method. When True, the datastream will be refreshed each update interval
    to ensure that data for the relevant contracts are being provided.

    When the data is 'static', the instrument being traded does not change
    over time. This is the more common scenario. In this case, the datastream
    is only refreshed during livetrading, to accomodate for new data coming in.
    In backtesting however, the entire dataset can be provided after the
    initial call, as it will not evolve during the backtest. Note that future
    data will not be provided to the strategy; instead, the data returned from
    the datastream will be filtered by each AutoTraderBot before being passed
    to the strategy.

    """

    def __init__(self, **kwargs):
        # Attributes
        self.instrument = None
        self.feed = None
        self.data_filepaths = None
        self.quote_data_file = None
        self.auxdata_files = None
        self.strategy_params = None
        self.get_data = None
        self.data_start = None
        self.data_end = None
        self.portfolio = None
        self.data_path_mapper = None

        # Unpack kwargs
        for item in kwargs:
            setattr(self, item, kwargs[item])

    def refresh(self, timestamp: datetime = None):
        """Returns up-to-date trading data for AutoBot to provide to the
        strategy.

        Parameters
        ----------
        timestamp : datetime, optional
            The current timestamp, which can be used to fetch
            data if need. Note that look-ahead is checked for in
            AutoTrader.autobot, so the data returned from this
            method can include all available data. The default is
            None.

        Returns
        -------
        data : pd.DataFrame
            The OHLC price data.
        multi_data : dict
            A dictionary of DataFrames.
        quote_data : pd.DataFrame
            The quote data.
        auxdata : dict
            Strategy auxiliary data.

        """
        # Retrieve main data
        if self.data_filepaths is not None:
            # Local data filepaths provided
            if isinstance(self.data_filepaths, str):
                # Single data filepath provided
                data = self.get_data._local(
                    self.data_filepaths, self.data_start, self.data_end
                )
                multi_data = None

            elif isinstance(self.data_filepaths, dict):
                # Multiple data filepaths provided
                multi_data = {}
                if self.portfolio:
                    for instrument, filepath in self.data_filepaths.items():
                        data = self.get_data._local(
                            filepath, self.data_start, self.data_end
                        )
                        multi_data[instrument] = data
                else:
                    for granularity, filepath in self.data_filepaths.items():
                        data = self.get_data._local(
                            filepath, self.data_start, self.data_end
                        )
                        multi_data[granularity] = data

                # Extract first dataset as base data (arbitrary)
                data = multi_data[list(self.data_filepaths.keys())[0]]

        elif self.data_path_mapper is not None:
            # Local data paths provided through mapper function
            multi_data = {}
            if self.portfolio:
                # Portfolio strategy
                for instrument in self.portfolio:
                    # Construct filepath
                    filepath = self.data_path_mapper(instrument)

                    # Save to multidata dict
                    data = self.get_data._local(
                        filepath, self.data_start, self.data_end
                    )
                    multi_data[instrument] = data

            else:
                # Single instrument strategy
                filepath = self.data_path_mapper(self.instrument)
                granularity = self.strategy_params["granularity"]
                data = self.get_data._local(filepath, self.data_start, self.data_end)
                multi_data = None

        else:
            # Download data
            multi_data = {}
            data_func = getattr(self.get_data, f"_{self.feed.lower()}")
            if self.portfolio:
                # Portfolio strategy
                if len(self.portfolio) > 1:
                    granularity = self.strategy_params["granularity"]
                    data_key = self.portfolio[0]
                    for instrument in self.portfolio:
                        data = data_func(
                            instrument,
                            granularity=granularity,
                            count=self.strategy_params["period"],
                            start_time=self.data_start,
                            end_time=self.data_end,
                        )
                        multi_data[instrument] = data
                else:
                    raise Exception(
                        "Portfolio strategies require more "
                        + "than a single instrument. Please set "
                        + "portfolio to False, or specify more "
                        + "instruments in the watchlist."
                    )
            else:
                # Single instrument strategy
                granularities = self.strategy_params["granularity"].split(",")
                data_key = granularities[0]
                for granularity in granularities:
                    data = data_func(
                        self.instrument,
                        granularity=granularity,
                        count=self.strategy_params["period"],
                        start_time=self.data_start,
                        end_time=self.data_end,
                    )
                    multi_data[granularity] = data

            # Take data as first element of multi-data
            data = multi_data[data_key]

            if len(multi_data) == 1:
                multi_data = None

        # Retrieve quote data
        if self.quote_data_file is not None:
            if isinstance(self.quote_data_file, str):
                # Single quote datafile
                quote_data = self.get_data._local(
                    self.quote_data_file, self.data_start, self.data_end
                )

            elif isinstance(quote_data, dict) and self.portfolio:
                # Multiple quote datafiles provided
                # TODO - support multiple quote data files (portfolio strategies)
                raise NotImplementedError(
                    "Locally-provided quote data not " + "implemented for portfolios."
                )
                quote_data = {}
                for instrument, path in quote_data.items():
                    quote_data[instrument] = self.get_data._local(
                        self.quote_data_file,  # need to specify
                        self.data_start,
                        self.data_end,
                    )
            else:
                raise Exception("Error in quote data file provided.")

        else:
            # Download data
            quote_data_func = getattr(self.get_data, f"_{self.feed.lower()}_quote_data")
            if self.portfolio:
                # Portfolio strategy - quote data for each instrument
                granularity = self.strategy_params["granularity"]
                quote_data = {}
                for instrument in self.portfolio:
                    quote_df = quote_data_func(
                        multi_data[instrument],
                        instrument,
                        granularity,
                        self.data_start,
                        self.data_end,
                        count=self.strategy_params["period"],
                    )
                    quote_data[instrument] = quote_df

            else:
                # Single instrument strategy - quote data for base granularity
                quote_data = quote_data_func(
                    data,
                    self.instrument,
                    self.strategy_params["granularity"].split(",")[0],
                    self.data_start,
                    self.data_end,
                    count=self.strategy_params["period"],
                )

        # Retrieve auxiliary data
        if self.auxdata_files is not None:
            if isinstance(self.auxdata_files, str):
                # Single data filepath provided
                auxdata = self.get_data._local(
                    self.auxdata_files, self.data_start, self.data_end
                )

            elif isinstance(self.auxdata_files, dict):
                # Multiple data filepaths provided
                auxdata = {}
                for key, filepath in self.auxdata_files.items():
                    data = self.get_data._local(
                        filepath, self.data_start, self.data_end
                    )
                    auxdata[key] = data
        else:
            auxdata = None

        # Correct any data mismatches
        if self.portfolio:
            # Portfolio strategy
            for instrument in multi_data:
                matched_data, matched_quote_data = self.match_quote_data(
                    multi_data[instrument], quote_data[instrument]
                )
                multi_data[instrument] = matched_data
                quote_data[instrument] = matched_quote_data
        else:
            # Single instrument data strategy
            if data is not None:
                # Data is not None (in case of 'none' data feed)
                data, quote_data = self.match_quote_data(data, quote_data)

        return data, multi_data, quote_data, auxdata

    def match_quote_data(
        self, data: pd.DataFrame, quote_data: pd.DataFrame
    ) -> pd.DataFrame:
        """Function to match index of trading data and quote data."""
        datasets = [data, quote_data]
        adjusted_datasets = []

        for dataset in datasets:
            # Initialise common index
            common_index = dataset.index

            # Update common index by intersection with other data
            for other_dataset in datasets:
                common_index = common_index.intersection(other_dataset.index)

            # Adjust data using common index found
            adj_data = dataset[dataset.index.isin(common_index)]

            adjusted_datasets.append(adj_data)

        # Unpack adjusted datasets
        adj_data, adj_quote_data = adjusted_datasets

        return adj_data, adj_quote_data

    def get_trading_bars(
        self,
        data: pd.DataFrame,
        quote_bars: bool,
        timestamp: datetime = None,
        processed_strategy_data: dict = None,
    ) -> dict:
        """Returns a dictionary of the current bars of the products being
        traded, based on the up-to-date data passed from autobot.

        Parameters
        ----------
        data : pd.DataFrame
            The strategy base OHLC data.
        quote_bars : bool
            Boolean flag to signal that quote data bars are being requested.
        processed_strategy_data : dict
            A dictionary containing all of the processed strategy data,
            allowing flexibility in what bars are returned.

        Returns
        -------
        dict
            A dictionary of OHLC bars, keyed by the product name.

        Notes
        -----
        The quote data bars dictionary must have the exact same keys as the
        trading bars dictionary. The quote_bars boolean flag is provided in
        case a distinction must be made when this method is called.
        """
        bars = {}
        strat_data = (
            processed_strategy_data["base"]
            if "base" in processed_strategy_data
            else processed_strategy_data
        )
        if isinstance(strat_data, dict):
            for instrument, data in strat_data.items():
                bars[instrument] = data.iloc[-1]
        else:
            bars[self.instrument] = strat_data.iloc[-1]

        return bars


class TradeWatcher:
    """Watches trade snapshots to detect new trades."""

    def __init__(self) -> None:
        self.last_trade_time = None
        self.latest_trades = []

    def update(self, trades):
        """Updates the trades being monitored for change."""
        if trades[0]["time"] != self.last_trade_time:
            # Trade update
            self.last_trade_time = trades[0]["time"]

            for trade in trades:
                if trade["time"] != self.last_trade_time:
                    break
                self.latest_trades.append(trade)

    def get_latest_trades(self):
        """Returns the latest (unseen) trades."""
        latest_trades = self.latest_trades
        self.latest_trades = []
        return latest_trades
